# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/MomentaryModeObserver.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1807 bytes
from . import Defaults


class MomentaryModeObserver:
    def __init__(self):
        object.__init__(self)
        self._controls = None
        self._mode_callback = None
        self._reset()

    def disconnect(self):
        self._reset()

    def set_mode_details(self, base_mode, controls, mode_callback=None):
        self._reset()
        if controls is not None:
            self._controls = controls
            for control in self._controls:
                control.add_value_listener(self._control_changed)

        self._base_mode = base_mode
        self._mode_callback = mode_callback

    def is_mode_momentary(self):
        return self._controls_changed or self._timer_count >= Defaults.MOMENTARY_DELAY_TICKS

    def on_timer(self):
        self._timer_count += 1

    def _control_changed(self, value):
        if self._mode_callback is None or self._mode_callback() == self._base_mode:
            self._controls_changed = True

    def _release_controls(self):
        if self._controls is not None:
            for control in self._controls:
                control.remove_value_listener(self._control_changed)

            self._controls = None

    def _reset(self):
        self._base_mode = -1
        self._controls_changed = False
        self._mode_callback = None
        self._timer_count = 0
        self._release_controls()



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class FloatVector:
    """
    A simple container for returning floats from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (FloatVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<float, std::__1::allocator<float>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (FloatVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<float, std::__1::allocator<float>> {lvalue},boost::python::api::object)
        """


class IntU64Vector:
    """
    A simple container for returning unsigned long integers from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (IntU64Vector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<unsigned long long, std::__1::allocator<unsigned long long>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (IntU64Vector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<unsigned long long, std::__1::allocator<unsigned long long>> {lvalue},boost::python::api::object)
        """


class IntVector:
    """
    A simple container for returning integers from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (IntVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<int, std::__1::allocator<int>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (IntVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<int, std::__1::allocator<int>> {lvalue},boost::python::api::object)
        """


class LimitationError:
    """
    Common base class for all non-exit exceptions.
    """

    def with_traceback(self) -> Any:
        """
        Exception.with_traceback(tb) --
        set self.__traceback__ to tb and return self.
        """

    args = None


class ObjectVector:
    """
    A simple read only container for returning python objects.
    """

    def append(self, arg2: object) -> None:
        """
        append( (ObjectVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<boost::python::api::object, std::__1::allocator<boost::python::api::object>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (ObjectVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<boost::python::api::object, std::__1::allocator<boost::python::api::object>> {lvalue},boost::python::api::object)
        """


class StringVector:
    """
    A simple container for returning strings from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (StringVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<TString, std::__1::allocator<TString>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (StringVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<TString, std::__1::allocator<TString>> {lvalue},boost::python::api::object)
        """


class Text:
    """
    A translatable, immutable string.
    """

    @property
    def text(self) -> Any:
        pass


class Timer:
    """
    A timer that will trigger a callback after a certain inverval. The timer can be repeated and will trigger the callback every interval. Errors in the callback will stop the timer.
    """

    def restart(self) -> None:
        """
        restart( (Timer)arg1) -> None :

            C++ signature :
                void restart(PythonTimer {lvalue})
        """

    def start(self) -> None:
        """
        start( (Timer)arg1) -> None :

            C++ signature :
                void start(PythonTimer {lvalue})
        """

    def stop(self) -> None:
        """
        stop( (Timer)arg1) -> None :

            C++ signature :
                void stop(PythonTimer {lvalue})
        """

    @property
    def running(self) -> Any:
        pass


class Vector:
    """
    A simple read only container for returning objects from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (Vector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<TWeakPtr<TPyHandleBase>, std::__1::allocator<TWeakPtr<TPyHandleBase>>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (Vector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<TWeakPtr<TPyHandleBase>, std::__1::allocator<TWeakPtr<TPyHandleBase>>> {lvalue},boost::python::api::object)
        """


def get_text(classname: str, textname: str) -> Text:
    """
    get_text( (str)classname, (str)textname) -> Text :
        Retrieves the (translated) Text identified by `classname` and `textname`.

        C++ signature :
            TText const* get_text(std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>,std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>)
    """


def log(arg1: str) -> None:
    """
    log( (str)arg1) -> None :

        C++ signature :
            void log(std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>)
    """


def subst_args(arg1: str = "", arg2: str = "", arg3: str = "", arg4: str = "", arg5: str = "") -> str:
    """
    subst_args( (Text)text [, (str)arg1='' [, (str)arg2='' [, (str)arg3='' [, (str)arg4='' [, (str)arg5='']]]]]) -> str :

        C++ signature :
            std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>> subst_args(TText [,std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>='' [,std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>='' [,std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>='' [,std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>='' [,std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>>='']]]]])
    """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ChannelTranslationSelector.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2426 bytes
from .ModeSelectorComponent import ModeSelectorComponent


class ChannelTranslationSelector(ModeSelectorComponent):
    def __init__(self, num_modes=0, *a, **k):
        (super().__init__)(*a, **k)
        self._controls_to_translate = None
        self._initial_num_modes = num_modes

    def disconnect(self):
        ModeSelectorComponent.disconnect(self)
        self._controls_to_translate = None

    def set_controls_to_translate(self, controls):
        for control in controls:
            pass

        self._controls_to_translate = controls

    def number_of_modes(self):
        result = self._initial_num_modes
        if result == 0:
            if self._modes_buttons is not None:
                result = len(self._modes_buttons)
        return result

    def update(self):
        super().update()
        if self._controls_to_translate is not None:
            for control in self._controls_to_translate:
                control.use_default_message()
                if self.is_enabled():
                    control.set_channel((control.message_channel() + self._mode_index) % 16)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/consts.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 771 bytes
from ableton.v2.control_surface import DEFAULT_PRIORITY

LOW_PRIORITY = DEFAULT_PRIORITY - 1
HIGH_PRIORITY = DEFAULT_PRIORITY + 1
M4L_PRIORITY = HIGH_PRIORITY + 1
MOMENTARY_DELAY = 0.3
ACTIVE_PARAMETER_TIMEOUT = 0.75



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4851 bytes
from .banking_util import (
    BANK_FORMAT,
    BANK_MAIN_KEY,
    BANK_PARAMETERS_KEY,
    MX_MAIN_BANK_INDEX,
    BankingInfo,
    all_parameters,
    device_bank_definition,
)
from .clip_creator import ClipCreator
from .component import Component
from .compound_element import CompoundElement, NestedElementClient
from .control_element import (
    ControlElement,
    ControlElementClient,
    ElementOwnershipHandler,
    NotifyingControlElement,
    OptimizedOwnershipHandler,
    get_element,
)
from .control_surface import ControlSurface, SimpleControlSurface
from .decoration import DecoratorFactory, LiveObjectDecorator, NotifyingList, PitchParameter, get_parameter_by_name
from .device_bank_registry import DeviceBankRegistry
from .device_chain_utils import find_instrument_devices, find_instrument_meeting_requirement
from .device_decorator_factory import DeviceDecoratorFactory
from .device_parameter_bank import DescribedDeviceParameterBank, create_device_bank
from .device_provider import DeviceProvider, device_to_appoint, select_and_appoint_device
from .identifiable_control_surface import IdentifiableControlSurface
from .input_control_element import (
    MIDI_CC_TYPE,
    MIDI_INVALID_TYPE,
    MIDI_NOTE_TYPE,
    MIDI_PB_TYPE,
    MIDI_SYSEX_TYPE,
    InputControlElement,
    InputSignal,
    ParameterSlot,
)
from .internal_parameter import (
    EnumWrappingParameter,
    IntegerParameter,
    InternalParameter,
    InternalParameterBase,
    RelativeInternalParameter,
    WrappingParameter,
    to_percentage_display,
)
from .layer import (
    BackgroundLayer,
    CompoundLayer,
    Layer,
    LayerClient,
    LayerError,
    SimpleLayerOwner,
    UnhandledElementError,
)
from .message_scheduler import MessageScheduler
from .midi_map import MidiMap
from .parameter_provider import ParameterInfo, ParameterProvider, is_parameter_quantized
from .parameter_slot_description import use
from .percussion_instrument_finder import PercussionInstrumentFinder
from .resource import (
    DEFAULT_PRIORITY,
    CompoundResource,
    ExclusiveResource,
    PrioritizedResource,
    ProxyResource,
    Resource,
    SharedResource,
    StackingResource,
)
from .session_ring_selection_linking import SessionRingSelectionLinking
from .simpler_decoration import BoolWrappingParameter, SimplerDeviceDecorator
from .skin import Skin, SkinColorMissingError, merge_skins
from .wavetable_decoration import (
    WavetableDeviceDecorator,
    WavetableEnvelopeType,
    WavetableFilterType,
    WavetableLfoType,
    WavetableOscillatorType,
)

__all__ = (
    "BackgroundLayer",
    "BANK_FORMAT",
    "BANK_MAIN_KEY",
    "BANK_PARAMETERS_KEY",
    "BankingInfo",
    "BoolWrappingParameter",
    "ClipCreator",
    "Component",
    "CompoundElement",
    "CompoundLayer",
    "CompoundResource",
    "ControlElement",
    "ControlElementClient",
    "ControlSurface",
    "DecoratorFactory",
    "DEFAULT_PRIORITY",
    "DescribedDeviceParameterBank",
    "DeviceBankRegistry",
    "DeviceDecoratorFactory",
    "DeviceProvider",
    "ElementOwnershipHandler",
    "EnumWrappingParameter",
    "ExclusiveResource",
    "IdentifiableControlSurface",
    "InputControlElement",
    "InputSignal",
    "IntegerParameter",
    "InternalParameter",
    "InternalParameterBase",
    "Layer",
    "LayerClient",
    "LayerError",
    "LiveObjectDecorator",
    "MessageScheduler",
    "MIDI_CC_TYPE",
    "MIDI_INVALID_TYPE",
    "MIDI_NOTE_TYPE",
    "MIDI_PB_TYPE",
    "MIDI_SYSEX_TYPE",
    "MidiMap",
    "MX_MAIN_BANK_INDEX",
    "NestedElementClient",
    "NotifyingControlElement",
    "NotifyingList",
    "OptimizedOwnershipHandler",
    "ParameterInfo",
    "ParameterProvider",
    "ParameterSlot",
    "PercussionInstrumentFinder",
    "PitchParameter",
    "PrioritizedResource",
    "ProxyResource",
    "RelativeInternalParameter",
    "Resource",
    "SessionRingSelectionLinking",
    "SharedResource",
    "SimpleControlSurface",
    "SimpleLayerOwner",
    "SimplerDeviceDecorator",
    "Skin",
    "SkinColorMissingError",
    "StackingResource",
    "UnhandledElementError",
    "WavetableDeviceDecorator",
    "WavetableEnvelopeType",
    "WavetableFilterType",
    "WavetableLfoType",
    "WavetableOscillatorType",
    "WrappingParameter",
    "all_parameters",
    "create_device_bank",
    "device_bank_definition",
    "device_to_appoint",
    "find_instrument_devices",
    "find_instrument_meeting_requirement",
    "get_element",
    "get_parameter_by_name",
    "is_parameter_quantized",
    "merge_skins",
    "select_and_appoint_device",
    "to_percentage_display",
    "use",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/control_surface_mapping.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4337 bytes
from ..base import Disconnectable
from .component_map import ComponentMap
from .layer import Layer
from .mode import EnablingAddLayerMode, ModesComponent


class ControlSurfaceMappingMixin(Disconnectable):
    def __init__(self, specification=None, *a, **k):
        self.component_map = ComponentMap(specification)
        (super().__init__)(*a, **k)

    def disconnect(self):
        super().disconnect()
        self.component_map = None

    def setup(self):
        self.component_map["Background"] = self._background
        self.component_map["Target_Track"] = self._target_track
        mappings = self.specification.create_mappings_function(self)
        component_names = self.component_map.keys()
        for name in list(mappings.keys()):
            if name in component_names:
                self._create_component(name, mappings.pop(name))

        for name, section in mappings.items():
            if name not in component_names:
                self._create_modes_component(name, section)

        for name, section in mappings.items():
            self._setup_modes_component(name, section)

    def _create_component(self, name, component_mappings):
        should_enable = component_mappings.pop("enable", True)
        component = self.component_map[name]
        component.layer = Layer(**component_mappings)
        component.set_enabled(should_enable)

    def _create_modes_component(self, name, modes_config):
        modes_component_type = modes_config.pop("modes_component_type", ModesComponent)
        component = modes_component_type(
            name=name,
            is_enabled=False,
            is_private=(modes_config.pop("is_private", False)),
            default_behaviour=(modes_config.pop("default_behaviour", None)),
            support_momentary_mode_cycling=(modes_config.pop("support_momentary_mode_cycling", True)),
        )
        self.component_map[name] = component

    def _setup_modes_component(self, name, modes_config):
        should_enable = modes_config.pop("enable", True)
        component = self.component_map[name]
        mode_control_layer = {}
        for mode_or_control_name, mode_or_element in modes_config.items():
            if isinstance(mode_or_element, str):
                mode_control_layer[mode_or_control_name] = mode_or_element
                continue
            else:
                self._add_mode(mode_or_control_name, mode_or_element, component)

        component.layer = Layer(**mode_control_layer)
        if component.selected_mode is None:
            component.selected_mode = component.modes[0]
        component.set_enabled(should_enable)

    def _add_mode(self, mode_name, mode_spec, modes_component):
        is_dict = isinstance(mode_spec, dict)
        behaviour = mode_spec.pop("behaviour", None) if is_dict else None
        selector = mode_spec.pop("selector", None) if is_dict else None
        mode = mode_spec
        if is_dict:
            if "modes" in mode_spec:
                mode = [self._create_mode_part(m) for m in mode_spec.pop("modes")]
            else:
                mode = self._create_mode_part(mode_spec)
        modes_component.add_mode(mode_name, mode, behaviour=behaviour, selector=selector)

    def _create_mode_part(self, mode_mappings):
        if isinstance(mode_mappings, dict):
            component = self.component_map[mode_mappings.pop("component")]
            if mode_mappings:
                return EnablingAddLayerMode(component=component, layer=Layer(**mode_mappings))
            return component
        return mode_mappings



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/simpler_decoration.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 14421 bytes
from functools import partial
from math import ceil, floor

import Live
from base.collection import IndexedDict
from past.builtins import unicode
from past.utils import old_div

from ..base import EventObject, clamp, listenable_property, listens, liveobj_valid, sign
from .decoration import LiveObjectDecorator
from .internal_parameter import (
    EnumWrappingParameter,
    IntegerParameter,
    RelativeInternalParameter,
    WrappingParameter,
    to_percentage_display,
)

BoolWrappingParameter = partial(
    WrappingParameter,
    to_property_value=(lambda integer, _simpler: bool(integer)),
    from_property_value=(lambda boolean, _simpler: int(boolean)),
    value_items=["Off", "On"],
    display_value_conversion=(lambda val: "On" if val else "Off"),
)


def from_user_range(minv, maxv):
    return lambda v, s: old_div(v - minv, float(maxv - minv))


def to_user_range(minv, maxv):
    return lambda v, s: clamp(v * (maxv - minv) + minv, minv, maxv)


def to_user_range_quantized(minv, maxv):
    user_range_transform = to_user_range(minv, maxv)
    return lambda v, s: int(round(user_range_transform(v, s)))


def from_sample_count(value, sample):
    return old_div(float(value), sample.length)


def to_sample_count(prev_value_getter, value, sample):
    truncation_func = ceil if sign(value - prev_value_getter()) > 0 else floor
    return clamp(int(truncation_func(value * sample.length)), 0, sample.length - 1)


SimplerWarpModes = IndexedDict(
    (
        (Live.Clip.WarpMode.beats, "Beats"),
        (Live.Clip.WarpMode.tones, "Tones"),
        (Live.Clip.WarpMode.texture, "Texture"),
        (Live.Clip.WarpMode.repitch, "Re-Pitch"),
        (Live.Clip.WarpMode.complex, "Complex"),
        (Live.Clip.WarpMode.complex_pro, "Pro"),
    ),
)


class SimplerDeviceDecorator(EventObject, LiveObjectDecorator):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._sample_based_parameters = []
        self._additional_parameters = []
        self.setup_parameters()
        self.register_disconnectables(self._decorated_parameters())
        self._SimplerDeviceDecorator__on_playback_mode_changed.subject = self._live_object
        self._SimplerDeviceDecorator__on_sample_changed.subject = self._live_object
        self._SimplerDeviceDecorator__on_slices_changed.subject = self._live_object.sample

    def setup_parameters(self):
        self.start = WrappingParameter(
            name="Start",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="start_marker",
            from_property_value=from_sample_count,
            to_property_value=(
                partial(
                    to_sample_count,
                    lambda: self.start.linear_value,
                )
            ),
        )
        self.end = WrappingParameter(
            name="End",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="end_marker",
            from_property_value=from_sample_count,
            to_property_value=(
                partial(
                    to_sample_count,
                    lambda: self.end.linear_value,
                )
            ),
        )
        self.sensitivity = WrappingParameter(
            name="Sensitivity",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="slicing_sensitivity",
            display_value_conversion=to_percentage_display,
        )
        self.mode = EnumWrappingParameter(
            name="Mode",
            parent=self,
            values_host=self,
            index_property_host=self,
            values_property="available_playback_modes",
            index_property="playback_mode",
        )
        self.slicing_playback_mode_param = EnumWrappingParameter(
            name="Playback",
            parent=self,
            values_host=self,
            index_property_host=self,
            values_property="available_slicing_playback_modes",
            index_property="slicing_playback_mode",
        )
        self.pad_slicing_param = BoolWrappingParameter(
            name="Pad Slicing",
            parent=self,
            property_host=(self._live_object),
            source_property="pad_slicing",
        )
        self.nudge = RelativeInternalParameter(name="Nudge", parent=self)
        self.multi_sample_mode_param = BoolWrappingParameter(
            name="Multi Sample",
            parent=self,
            property_host=(self._live_object),
            source_property="multi_sample_mode",
        )
        self.warp = BoolWrappingParameter(
            name="Warp",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="warping",
        )
        self.warp_mode_param = EnumWrappingParameter(
            name="Warp Mode",
            parent=self,
            values_host=self,
            index_property_host=(self._live_object.sample),
            values_property="available_warp_modes",
            index_property="warp_mode",
            to_index_conversion=(lambda i: Live.Clip.WarpMode(SimplerWarpModes.key_by_index(i))),
            from_index_conversion=(lambda i: SimplerWarpModes.index_by_key(i)),
        )
        self.voices_param = EnumWrappingParameter(
            name="Voices",
            parent=self,
            values_host=self,
            index_property_host=self,
            values_property="available_voice_numbers",
            index_property="voices",
            to_index_conversion=(lambda i: self.available_voice_numbers[i]),
            from_index_conversion=(lambda i: self.available_voice_numbers.index(i)),
        )
        self.pitch_bend_range_param = IntegerParameter(
            name="PB Range",
            parent=self,
            integer_value_host=(self._live_object),
            integer_value_property_name="pitch_bend_range",
            min_value=0,
            max_value=24,
            show_as_quantized=False,
            display_value_conversion=(lambda x: unicode(x) + " st"),
        )
        self.note_pitch_bend_range_param = IntegerParameter(
            name="Note PB",
            parent=self,
            integer_value_host=(self._live_object),
            integer_value_property_name="note_pitch_bend_range",
            min_value=0,
            max_value=48,
            show_as_quantized=False,
            display_value_conversion=(lambda x: unicode(x) + " st"),
        )
        self.granulation_resolution = EnumWrappingParameter(
            name="Preserve",
            parent=self,
            values_host=self,
            index_property_host=(self._live_object.sample),
            values_property="available_resolutions",
            index_property="beats_granulation_resolution",
        )
        self.transient_loop_mode = EnumWrappingParameter(
            name="Loop Mode",
            parent=self,
            values_host=self,
            index_property_host=(self._live_object.sample),
            values_property="available_transient_loop_modes",
            index_property="beats_transient_loop_mode",
        )
        self.transient_envelope = WrappingParameter(
            name="Envelope",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="beats_transient_envelope",
            from_property_value=(from_user_range(0.0, 100.0)),
            to_property_value=(to_user_range(0.0, 100.0)),
        )
        self.tones_grain_size_param = WrappingParameter(
            name="Grain Size Tones",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="tones_grain_size",
            from_property_value=(from_user_range(12.0, 100.0)),
            to_property_value=(to_user_range(12.0, 100.0)),
        )
        self.texture_grain_size_param = WrappingParameter(
            name="Grain Size Texture",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="texture_grain_size",
            from_property_value=(from_user_range(2.0, 263.0)),
            to_property_value=(to_user_range(2.0, 263.0)),
        )
        self.flux = WrappingParameter(
            name="Flux",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="texture_flux",
            from_property_value=(from_user_range(0.0, 100.0)),
            to_property_value=(to_user_range(0.0, 100.0)),
        )
        self.formants = WrappingParameter(
            name="Formants",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="complex_pro_formants",
            from_property_value=(from_user_range(0.0, 100.0)),
            to_property_value=(to_user_range(0.0, 100.0)),
        )
        self.complex_pro_envelope_param = WrappingParameter(
            name="Envelope Complex Pro",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="complex_pro_envelope",
            from_property_value=(from_user_range(8.0, 256.0)),
            to_property_value=(to_user_range(8.0, 256.0)),
        )
        self.gain_param = WrappingParameter(
            name="Gain",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="gain",
            display_value_conversion=(
                lambda _: self._live_object.sample.gain_display_string()
                if liveobj_valid(self._live_object) and liveobj_valid(self._live_object.sample)
                else ""
            ),
        )
        self.slicing_style_param = EnumWrappingParameter(
            name="Slice by",
            parent=self,
            values_host=self,
            index_property_host=(self._live_object.sample),
            values_property="available_slice_styles",
            index_property="slicing_style",
        )
        self.slicing_beat_division_param = EnumWrappingParameter(
            name="Division",
            parent=self,
            values_host=self,
            index_property_host=(self._live_object.sample),
            values_property="available_slicing_beat_divisions",
            index_property="slicing_beat_division",
        )
        self.slicing_region_count_param = WrappingParameter(
            name="Regions",
            parent=self,
            property_host=(self._live_object.sample),
            source_property="slicing_region_count",
            from_property_value=(from_user_range(2, 64)),
            to_property_value=(to_user_range_quantized(2, 64)),
        )
        self._sample_based_parameters.extend(
            [
                self.start,
                self.end,
                self.sensitivity,
                self.warp,
                self.transient_envelope,
                self.tones_grain_size_param,
                self.texture_grain_size_param,
                self.flux,
                self.formants,
                self.complex_pro_envelope_param,
                self.gain_param,
                self.slicing_region_count_param,
                self.warp_mode_param,
                self.granulation_resolution,
                self.transient_loop_mode,
                self.slicing_style_param,
                self.slicing_beat_division_param,
            ],
        )
        self._additional_parameters.extend(
            [
                self.mode,
                self.slicing_playback_mode_param,
                self.pad_slicing_param,
                self.nudge,
                self.multi_sample_mode_param,
                self.voices_param,
                self.pitch_bend_range_param,
                self.note_pitch_bend_range_param,
            ],
        )

    def _decorated_parameters(self):
        return tuple(self._sample_based_parameters) + tuple(self._additional_parameters)

    @property
    def parameters(self):
        return tuple(self._live_object.parameters) + self._decorated_parameters()

    @property
    def available_playback_modes(self):
        return ["Classic", "One-Shot", "Slicing"]

    @property
    def available_slicing_playback_modes(self):
        return ["Mono", "Poly", "Thru"]

    @property
    def available_voice_numbers(self):
        return list(Live.SimplerDevice.get_available_voice_numbers())

    @property
    def available_warp_modes(self):
        return list(SimplerWarpModes.values())

    @property
    def available_resolutions(self):
        return ("1 Bar", "1/2", "1/4", "1/8", "1/16", "1/32", "Transients")

    @property
    def available_slice_styles(self):
        return ("Transient", "Beat", "Region", "Manual")

    @property
    def available_slicing_beat_divisions(self):
        return ("1/16", "1/16T", "1/8", "1/8T", "1/4", "1/4T", "1/2", "1/2T", "1 Bar", "2 Bars", "4 Bars")

    @property
    def available_transient_loop_modes(self):
        return ("Off", "Forward", "Alternate")

    @listenable_property
    def current_playback_mode(self):
        return self._live_object.playback_mode

    @listenable_property
    def slices(self):
        if liveobj_valid(self._live_object):
            if liveobj_valid(self._live_object.sample):
                return self._live_object.sample.slices
        return []

    @listens("sample")
    def __on_sample_changed(self):
        self._on_sample_changed()

    def _on_sample_changed(self):
        self._reconnect_sample_listeners()

    def _reconnect_sample_listeners(self):
        for param in self._sample_based_parameters:
            param.set_property_host(self._live_object.sample)

        self._reconnect_to_slices()

    def _reconnect_to_slices(self):
        self._SimplerDeviceDecorator__on_slices_changed.subject = self._live_object.sample
        self.notify_slices()

    @listens("slices")
    def __on_slices_changed(self):
        self._on_slices_changed()

    def _on_slices_changed(self):
        self.notify_slices()

    @listens("playback_mode")
    def __on_playback_mode_changed(self):
        self.notify_current_playback_mode()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/simpler_slice_nudging.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4529 bytes
from contextlib import contextmanager

import Live
from past.utils import old_div

from ..base import EventObject, clamp, find_if, listens, liveobj_valid

CENTERED_NUDGE_VALUE = 0.5
MINIMUM_SLICE_DISTANCE = 2


def is_simpler(device):
    return device and device.class_name == "OriginalSimpler"


class SimplerSliceNudging(EventObject):
    _simpler = None
    _nudge_parameter = None

    def set_device(self, device):
        self._simpler = device if is_simpler(device) else None
        self._SimplerSliceNudging__on_selected_slice_changed.subject = self._simpler
        with self._updating_nudge_parameter():
            self._nudge_parameter = find_if(
                lambda p: p.name == "Nudge",
                self._simpler.parameters if liveobj_valid(self._simpler) else [],
            )

    @contextmanager
    def _updating_nudge_parameter(self):
        if self._nudge_parameter:
            self._nudge_parameter.set_display_value_conversion(None)
        yield
        if self._nudge_parameter:
            self._nudge_parameter.set_display_value_conversion(self._display_value_conversion)
        self._SimplerSliceNudging__on_nudge_delta.subject = self._nudge_parameter

    def _can_access_slicing_properties(self):
        return (
            liveobj_valid(self._simpler)
            and liveobj_valid(self._simpler.sample)
            and self._simpler.current_playback_mode == Live.SimplerDevice.PlaybackMode.slicing
        )

    @listens("view.selected_slice")
    def __on_selected_slice_changed(self):
        if self._nudge_parameter:
            self._nudge_parameter.notify_value()

    @listens("delta")
    def __on_nudge_delta(self, delta):
        if self._can_access_slicing_properties():
            old_slice_time = self._simpler.view.selected_slice
            if old_slice_time >= 0:
                if self._is_first_slice_at_time(old_slice_time):
                    new_start = self._new_start_marker_time(old_slice_time, delta)
                    self._simpler.sample.start_marker = new_start
                    return
                new_slice_time = old_slice_time + self._sample_change_from_delta(delta)
                if old_slice_time != new_slice_time:
                    original_slices = self._simpler.sample.slices
                    returned_time = self._simpler.sample.move_slice(old_slice_time, new_slice_time)
                    try:
                        self._simpler.view.selected_slice = returned_time
                    except RuntimeError:
                        self._simpler.view.selected_slice = self._simpler.slices[
                            list(original_slices).index(old_slice_time)
                        ]

    def _is_first_slice_at_time(self, slice_time):
        start_sample = self._simpler.sample.start_marker
        return abs(slice_time - start_sample) < MINIMUM_SLICE_DISTANCE

    def _new_start_marker_time(self, old_slice_time, delta):
        change_in_samples = self._sample_change_from_delta(delta)
        new_start_marker_time = old_slice_time + change_in_samples
        return clamp(new_start_marker_time, 0, self._simpler.sample.length - MINIMUM_SLICE_DISTANCE)

    def _sample_change_from_delta(self, delta):
        sample_length = self._simpler.sample.length
        change_in_samples = round(old_div(delta * sample_length, 10))
        return int(change_in_samples)

    def _display_value_conversion(self, _value):
        selected_slice = self._simpler.view.selected_slice if self._can_access_slicing_properties() else -1
        if selected_slice >= 0:
            return str(selected_slice)
        return "-"



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/DrumGroupComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 13359 bytes
from .Control import ButtonControl, PlayableControl, control_matrix
from .Dependency import depends
from .SlideComponent import Slideable, SlideComponent
from .SubjectSlot import subject_slot, subject_slot_group
from .Util import clamp, find_if, first

BASE_DRUM_RACK_NOTE = 36


class DrumGroupComponent(SlideComponent, Slideable):
    __subject_events__ = ("pressed_pads",)
    mute_button = ButtonControl()
    solo_button = ButtonControl()
    delete_button = ButtonControl()
    quantize_button = ButtonControl()
    select_button = ButtonControl()
    drum_matrix = control_matrix(PlayableControl)

    @depends(set_pad_translations=None)
    def __init__(self, translation_channel=None, set_pad_translations=None, *a, **k):
        self._takeover_drums = False
        self._drum_group_device = None
        self._selected_drum_pad = None
        self._all_drum_pads = []
        self._selected_pads = []
        self._visible_drum_pads = []
        self._translation_channel = translation_channel
        self._coordinate_to_pad_map = {}
        (super().__init__)(*a, **k)
        self._set_pad_translations = set_pad_translations

    position_count = 32
    page_length = 4
    page_offset = 1

    def contents_range(self, pmin, pmax):
        pos_count = self.position_count
        first_pos = max(int(pmin - 0.05), 0)
        last_pos = min(int(pmax + 0.2), pos_count)
        return list(range(first_pos, last_pos))

    def contents(self, index):
        drum = self._drum_group_device
        if drum:
            return any(pad.chains for pad in drum.drum_pads[index * 4 : index * 4 + 4])
        return False

    def _get_position(self):
        if self._drum_group_device:
            return self._drum_group_device.view.drum_pads_scroll_position
        return 0

    def _set_position(self, index):
        if self._drum_group_device:
            self._drum_group_device.view.drum_pads_scroll_position = index

    position = property(_get_position, _set_position)

    @property
    def width(self):
        if self.drum_matrix.width:
            return self.drum_matrix.width
        return 4

    @property
    def height(self):
        if self.drum_matrix.height:
            return self.drum_matrix.height
        return 4

    @property
    def pressed_pads(self):
        return self._selected_pads

    @property
    def visible_drum_pads(self):
        if self._visible_drum_pads:
            if self._all_drum_pads:
                first_pad = first(self._visible_drum_pads)
                if first_pad:
                    size = self.width * self.height
                    first_note = first_pad.note
                    if first_note > 128 - size:
                        size = 128 - first_note
                    offset = clamp(first_note, 0, 128 - len(self._visible_drum_pads))
                    return self._all_drum_pads[offset : offset + size]
        return []

    def update(self):
        super().update()
        self._set_control_pads_from_script(False)
        self._update_led_feedback()

    def set_drum_matrix(self, matrix):
        self.drum_matrix.set_control_element(matrix)
        for button in self.drum_matrix:
            button.channel = self._translation_channel

        if self._selected_pads:
            self._selected_pads = []
            self.notify_pressed_pads()
        self._create_and_set_pad_translations()
        self._update_control_from_script()
        self._update_identifier_translations()
        self._update_led_feedback()

    def set_drum_group_device(self, drum_group_device):
        if drum_group_device:
            if not drum_group_device.can_have_drum_pads:
                drum_group_device = None
            if drum_group_device != self._drum_group_device:
                self._on_visible_drum_pads_changed.subject = drum_group_device
                drum_group_view = drum_group_device.view if drum_group_device else None
                self._on_selected_drum_pad_changed.subject = drum_group_view
                self._on_drum_pads_scroll_position_changed.subject = drum_group_view
                self._drum_group_device = drum_group_device
                self._update_drum_pad_listeners()
                self._on_selected_drum_pad_changed()
                self._update_identifier_translations()
                super().update()

    def _update_drum_pad_listeners(self):
        if self._drum_group_device:
            self._all_drum_pads = self._drum_group_device.drum_pads
            self._visible_drum_pads = self._drum_group_device.visible_drum_pads
            self._on_solo_changed.replace_subjects(self._visible_drum_pads)
            self._on_mute_changed.replace_subjects(self._visible_drum_pads)
            self._update_identifier_translations()

    @subject_slot_group("solo")
    def _on_solo_changed(self, pad):
        self._update_led_feedback()

    @subject_slot_group("mute")
    def _on_mute_changed(self, pad):
        self._update_led_feedback()

    def _update_led_feedback(self):
        if self._drum_group_device:
            soloed_pads = find_if(lambda pad: pad.solo, self._all_drum_pads)
            for button in self.drum_matrix:
                pad = self._coordinate_to_pad_map.get(button.coordinate, None)
                if pad:
                    self._update_pad_led(pad, button, soloed_pads)

    def _update_pad_led(self, pad, button, soloed_pads):
        button_color = "DrumGroup.PadEmpty"
        if pad == self._selected_drum_pad:
            if soloed_pads and not pad.solo or pad.mute:
                button_color = "DrumGroup.PadSelectedNotSoloed"
            else:
                if pad.mute and not pad.solo:
                    button_color = "DrumGroup.PadMutedSelected"
                else:
                    if soloed_pads and pad.solo:
                        button_color = "DrumGroup.PadSoloedSelected"
                    else:
                        button_color = "DrumGroup.PadSelected"
        else:
            if pad.chains:
                if soloed_pads and not pad.solo:
                    button_color = "DrumGroup.PadFilled" if not pad.mute else "DrumGroup.PadMuted"
                else:
                    if not soloed_pads or pad.mute:
                        button_color = "DrumGroup.PadMuted"
                    else:
                        button_color = "DrumGroup.PadSoloed" if soloed_pads and pad.solo else "DrumGroup.PadFilled"
            else:
                button_color = "DrumGroup.PadEmpty"
        button.color = button_color

    def _button_coordinates_to_pad_index(self, first_note, coordinates):
        y, x = coordinates
        y = self.height - y - 1
        if x < 4 and y >= 4:
            first_note += 16
        else:
            if x >= 4 and y < 4:
                first_note += 4 * self.width
            else:
                if x >= 4:
                    if y >= 4:
                        first_note += 4 * self.width + 16
        return x % 4 + y % 4 * 4 + first_note

    @drum_matrix.pressed
    def drum_matrix(self, pad):
        self._on_matrix_pressed(pad)

    @drum_matrix.released
    def drum_matrix(self, pad):
        self._on_matrix_released(pad)

    def _on_matrix_released(self, pad):
        selected_drum_pad = self._coordinate_to_pad_map[pad.coordinate]
        if selected_drum_pad in self._selected_pads:
            self._selected_pads.remove(selected_drum_pad)
            if not self._selected_pads:
                self._update_control_from_script()
            self.notify_pressed_pads()
        self._update_led_feedback()

    def _on_matrix_pressed(self, pad):
        selected_drum_pad = self._coordinate_to_pad_map[pad.coordinate]
        if self.mute_button.is_pressed:
            selected_drum_pad.mute = not selected_drum_pad.mute
        if self.solo_button.is_pressed:
            selected_drum_pad.solo = not selected_drum_pad.solo
        if self.quantize_button.is_pressed:
            pad.color = "DrumGroup.PadAction"
            self.quantize_pitch(selected_drum_pad.note)
        if self.delete_button.is_pressed:
            pad.color = "DrumGroup.PadAction"
            self.delete_pitch(selected_drum_pad)
        if self.select_button.is_pressed:
            self._drum_group_device.view.selected_drum_pad = selected_drum_pad
            self.select_drum_pad(selected_drum_pad)
            self._selected_pads.append(selected_drum_pad)
            if len(self._selected_pads) == 1:
                self._update_control_from_script()
            self.notify_pressed_pads()
        if self.mute_button.is_pressed or self.solo_button.is_pressed:
            self._update_led_feedback()

    @subject_slot("visible_drum_pads")
    def _on_visible_drum_pads_changed(self):
        self._update_drum_pad_listeners()
        self._update_led_feedback()

    @subject_slot("drum_pads_scroll_position")
    def _on_drum_pads_scroll_position_changed(self):
        self._update_identifier_translations()
        self._update_led_feedback()
        self.notify_position()

    @subject_slot("selected_drum_pad")
    def _on_selected_drum_pad_changed(self):
        self._selected_drum_pad = self._drum_group_device.view.selected_drum_pad if self._drum_group_device else None
        self._update_led_feedback()

    @mute_button.value
    def mute_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @solo_button.value
    def solo_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @delete_button.value
    def delete_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @quantize_button.value
    def quantize_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @select_button.value
    def select_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    def _set_control_pads_from_script(self, takeover_drums):
        if takeover_drums != self._takeover_drums:
            self._takeover_drums = takeover_drums
            self._update_control_from_script()

    def _update_control_from_script(self):
        takeover_drums = self._takeover_drums or bool(self._selected_pads)
        for button in self.drum_matrix:
            button.set_playable(not takeover_drums)

    def _update_identifier_translations(self):
        if self.visible_drum_pads:
            if not self._can_set_pad_translations():
                self._set_non_pad_translated_identifiers()
            else:
                self._set_pad_translated_identifiers()

    def _set_non_pad_translated_identifiers(self):
        visible_drum_pads = self.visible_drum_pads
        if visible_drum_pads:
            first_pad = first(visible_drum_pads)
            for button in self.drum_matrix:
                identifier = self._button_coordinates_to_pad_index(first_pad.note, button.coordinate)
                if identifier < 128:
                    button.identifier = identifier
                    button.enabled = True
                    self._coordinate_to_pad_map[button.coordinate] = self._all_drum_pads[button.identifier]
                else:
                    button.enabled = False

    def _set_pad_translated_identifiers(self):
        visible_drum_pads = self.visible_drum_pads
        if visible_drum_pads:
            for index, button in enumerate(self.drum_matrix):
                row, col = button.coordinate
                self._coordinate_to_pad_map[(self.width - 1 - row, col)] = visible_drum_pads[index]

    def _can_set_pad_translations(self):
        return self.width <= 4 and self.height <= 4

    def _create_and_set_pad_translations(self):
        def create_translation_entry(button):
            row, col = button.coordinate
            button.identifier = self._button_coordinates_to_pad_index(BASE_DRUM_RACK_NOTE, button.coordinate)
            return (col, row, button.identifier, button.channel)

        if self._can_set_pad_translations():
            translations = tuple(map(create_translation_entry, self.drum_matrix))
            self._set_pad_translated_identifiers()
        else:
            translations = None
            self._set_non_pad_translated_identifiers()
        self._set_pad_translations(translations)

    def select_drum_pad(self, drum_pad):
        raise NotImplementedError

    def quantize_pitch(self, note):
        raise NotImplementedError

    def delete_pitch(self, drum_pad):
        raise NotImplementedError



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/session_overview.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5619 bytes
from ...base import depends, listens, listens_group
from .. import Component
from ..controls import ButtonControl, control_matrix


class SessionOverviewComponent(Component):
    matrix = control_matrix(ButtonControl, color=None)

    @depends(session_ring=None)
    def __init__(self, name="Session_Overview", session_ring=None, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._track_bank_offset = 0
        self._scene_bank_offset = 0
        self._track_bank_size = 0
        self._session_ring = session_ring
        self._SessionOverviewComponent__on_session_offset_changed.subject = self._session_ring
        self.register_slot(self._session_ring, self.update, "tracks")
        self.register_slot(self.song, self.update, "scenes")

    def set_matrix(self, matrix):
        self.matrix.set_control_element(matrix)
        self.update()

    @matrix.pressed
    def matrix(self, button):
        y, x = button.coordinate
        track_offset = x * self._session_ring.num_tracks + self._track_bank_offset
        scene_offset = y * self._session_ring.num_scenes + self._scene_bank_offset
        if track_offset in range(len(self._session_ring.tracks_to_use())):
            if scene_offset in range(len(self.song.scenes)):
                self._session_ring.set_offsets(track_offset, scene_offset)

    @listens("offset")
    def __on_session_offset_changed(self, *_):
        self.update()

    @listens_group("playing_slot_index")
    def __on_playing_slot_index_changed(self, _):
        self._update_matrix()

    def update(self):
        super().update()
        if self.is_enabled() and self.matrix.control_elements:
            self._update_bank_offsets()
            self._update_matrix()
            self._SessionOverviewComponent__on_playing_slot_index_changed.replace_subjects(
                self._session_ring.tracks_to_use()[
                    self._track_bank_offset : self._track_bank_offset + self._track_bank_size
                ],
            )
        else:
            self._SessionOverviewComponent__on_playing_slot_index_changed.replace_subjects([])

    def _block_is_within_selection(self, x, y, num_tracks, num_scenes):
        return self._session_ring.track_offset - self._track_bank_offset in range(
            num_tracks * (x - 1) + 1,
            num_tracks * (x + 1),
        ) and self._session_ring.scene_offset - self._scene_bank_offset in range(
            num_scenes * (y - 1) + 1,
            num_scenes * (y + 1),
        )

    def _block_has_playing_clips(self, tracks, num_tracks, num_scenes, track_offset, scene_offset):
        for track in tracks[track_offset : track_offset + num_tracks]:
            if track in self.song.tracks:
                if track.playing_slot_index in range(scene_offset, scene_offset + num_scenes):
                    return True

        return False

    def _update_matrix(self):
        num_tracks = self._session_ring.num_tracks
        num_scenes = self._session_ring.num_scenes
        tracks = self._session_ring.tracks_to_use()
        for x in range(self.matrix.width):
            for y in range(self.matrix.height):
                track_offset = x * num_tracks + self._track_bank_offset
                scene_offset = y * num_scenes + self._scene_bank_offset
                if track_offset not in range(len(tracks)) or scene_offset not in range(len(self.song.scenes)):
                    self.matrix.get_control(y, x).color = "Zooming.Empty"
                    continue
                if self._block_is_within_selection(x, y, num_tracks, num_scenes):
                    self.matrix.get_control(y, x).color = "Zooming.Selected"
                else:
                    if self._block_has_playing_clips(tracks, num_tracks, num_scenes, track_offset, scene_offset):
                        self.matrix.get_control(y, x).color = "Zooming.Playing"
                    else:
                        self.matrix.get_control(y, x).color = "Zooming.Stopped"

    def _update_bank_offsets(self):
        scene_bank_size = self.matrix.height * self._session_ring.num_scenes
        self._scene_bank_offset = self._session_ring.scene_offset // scene_bank_size * scene_bank_size
        self._track_bank_size = self.matrix.width * self._session_ring.num_tracks
        self._track_bank_offset = self._session_ring.track_offset // self._track_bank_size * self._track_bank_size



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ButtonSliderElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4760 bytes
from past.utils import old_div

from .InputControlElement import MIDI_INVALID_TYPE, InputControlElement
from .SliderElement import SliderElement
from .SubjectSlot import SlotManager


class ButtonSliderElement(SliderElement, SlotManager):
    _last_sent_value = -1

    def __init__(self, buttons):
        SliderElement.__init__(self, MIDI_INVALID_TYPE, 0, 0)
        self._parameter_value_slot = self.register_slot(None, self._on_parameter_changed, "value")
        self._buttons = buttons
        self._button_slots = self.register_slot_manager()
        for button in self._buttons:
            self._button_slots.register_slot(button, (self._button_value), "value", extra_kws={"identify_sender": True})

    def disconnect(self):
        SliderElement.disconnect(self)
        self._buttons = None

    def message_channel(self):
        raise NotImplementedError("message_channel() should not be called directly on ButtonSliderElement")

    def message_identifier(self):
        raise NotImplementedError("message_identifier() should not be called directly on ButtonSliderElement")

    def message_map_mode(self):
        raise NotImplementedError("message_map_mode() should not be called directly on ButtonSliderElement")

    def install_connections(self, install_translation_callback, install_mapping_callback, install_forwarding_callback):
        pass

    def connect_to(self, parameter):
        InputControlElement.connect_to(self, parameter)
        self._parameter_value_slot.subject = parameter
        if self._parameter_to_map_to is not None:
            self._on_parameter_changed()

    def release_parameter(self):
        self._parameter_value_slot.subject = None
        InputControlElement.release_parameter(self)

    def identifier_bytes(self):
        raise RuntimeWarning("identifier_bytes() should not be called on ButtonSliderElement")

    def send_value(self, value):
        if value != self._last_sent_value:
            num_buttons = len(self._buttons)
            index_to_light = 0
            index_to_light = int(old_div((num_buttons - 1) * value, 127)) if value > 0 else 0
            for index in range(num_buttons):
                if index == index_to_light:
                    self._buttons[index].turn_on()
                else:
                    self._buttons[index].turn_off()

            self._last_sent_value = value

    def _button_value(self, value, sender):
        self.clear_send_cache()
        if not (value != 0 or sender.is_momentary()):
            index_of_sender = list(self._buttons).index(sender)
            midi_value = int(old_div(127 * index_of_sender, len(self._buttons) - 1))
            if self._parameter_to_map_to is not None:
                if self._parameter_to_map_to.is_enabled:
                    param_range = self._parameter_to_map_to.max - self._parameter_to_map_to.min
                    param_value = (
                        old_div(param_range * index_of_sender, len(self._buttons) - 1) + self._parameter_to_map_to.min
                    )
                    if index_of_sender > 0:
                        param_value += old_div(param_range, 4 * len(self._buttons))
                        if param_value > self._parameter_to_map_to.max:
                            param_value = self._parameter_to_map_to.max
                    self._parameter_to_map_to.value = param_value
            self.notify_value(midi_value)

    def _on_parameter_changed(self):
        param_range = abs(self._parameter_to_map_to.max - self._parameter_to_map_to.min)
        midi_value = int(
            old_div(127 * abs(self._parameter_to_map_to.value - self._parameter_to_map_to.min), param_range),
        )
        self.send_value(midi_value)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/transport.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9125 bytes
from functools import partial

from past.utils import old_div

from ...base import clamp, in_range, listens, task
from ..component import Component
from ..control import ButtonControl, ToggleButtonControl
from .toggle import ToggleComponent

TEMPO_TOP = 200.0
TEMPO_BOTTOM = 60.0
TEMPO_FINE_RANGE = 2.56
SEEK_SPEED = 10.0


class TransportComponent(Component):
    play_button = ToggleButtonControl(toggled_color="Transport.PlayOn", untoggled_color="Transport.PlayOff")
    stop_button = ButtonControl()
    continue_playing_button = ButtonControl()
    tap_tempo_button = ButtonControl(color="DefaultButton.On", pressed_color="DefaultButton.Off")

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._ffwd_button = None
        self._rwd_button = None
        self._tap_tempo_button = None
        self._tempo_control = None
        self._tempo_fine_control = None
        self._rwd_task = task.Task()
        self._ffwd_task = task.Task()
        self._fine_tempo_needs_pickup = True
        self._prior_fine_tempo_value = -1
        self._end_undo_step_task = self._tasks.add(task.sequence(task.wait(1.5), task.run(self.song.end_undo_step)))
        self._end_undo_step_task.kill()
        song = self.song
        self._loop_toggle = ToggleComponent("loop", song, parent=self)
        self._punch_in_toggle = ToggleComponent("punch_in", song, is_momentary=True, parent=self)
        self._punch_out_toggle = ToggleComponent("punch_out", song, is_momentary=True, parent=self)
        self._record_toggle = ToggleComponent("record_mode", song, parent=self)
        self._nudge_down_toggle = ToggleComponent("nudge_down", song, is_momentary=True, parent=self)
        self._nudge_up_toggle = ToggleComponent("nudge_up", song, is_momentary=True, parent=self)
        self._metronome_toggle = ToggleComponent("metronome", song, parent=self)
        self._arrangement_overdub_toggle = ToggleComponent("arrangement_overdub", song, parent=self)
        self._overdub_toggle = ToggleComponent("overdub", song, parent=self)
        self._TransportComponent__on_is_playing_changed.subject = song
        self._TransportComponent__on_is_playing_changed()

    @continue_playing_button.pressed
    def continue_playing_button(self, _):
        song = self.song
        if not song.is_playing:
            song.continue_playing()
        else:
            song.stop_playing()

    @listens("is_playing")
    def __on_is_playing_changed(self):
        self._update_button_states()

    def _update_button_states(self):
        self.play_button.is_toggled = self.song.is_playing
        self._update_stop_button_color()

    def _update_stop_button_color(self):
        self.stop_button.color = (
            self.play_button.toggled_color if self.play_button.is_toggled else self.play_button.untoggled_color
        )

    @stop_button.released
    def _on_stop_button_released(self, button):
        self.song.is_playing = False

    @play_button.toggled
    def _on_play_button_toggled(self, is_toggled, button):
        self.song.is_playing = is_toggled

    def set_seek_buttons(self, ffwd_button, rwd_button):
        if self._ffwd_button != ffwd_button:
            self._ffwd_button = ffwd_button
            self._TransportComponent__ffwd_value_slot.subject = ffwd_button
            self._ffwd_task.kill()
        if self._rwd_button != rwd_button:
            self._rwd_button = rwd_button
            self._TransportComponent__rwd_value_slot.subject = rwd_button
            self._rwd_task.kill()

    def set_seek_forward_button(self, ffwd_button):
        if self._ffwd_button != ffwd_button:
            self._ffwd_button = ffwd_button
            self._TransportComponent__ffwd_value_slot.subject = ffwd_button
            self._ffwd_task.kill()

    def set_seek_backward_button(self, rwd_button):
        if self._rwd_button != rwd_button:
            self._rwd_button = rwd_button
            self._TransportComponent__rwd_value_slot.subject = rwd_button
            self._rwd_task.kill()

    def set_nudge_buttons(self, up_button, down_button):
        self._nudge_up_toggle.set_toggle_button(up_button)
        self._nudge_down_toggle.set_toggle_button(down_button)

    def set_nudge_up_button(self, up_button):
        self._nudge_up_toggle.set_toggle_button(up_button)

    def set_nudge_down_button(self, down_button):
        self._nudge_down_toggle.set_toggle_button(down_button)

    def set_record_button(self, button):
        self._record_toggle.set_toggle_button(button)

    def set_loop_button(self, button):
        self._loop_toggle.set_toggle_button(button)

    def set_punch_in_button(self, in_button):
        self._punch_in_toggle.set_toggle_button(in_button)

    def set_punch_out_button(self, out_button):
        self._punch_out_toggle.set_toggle_button(out_button)

    def set_metronome_button(self, button):
        self._metronome_toggle.set_toggle_button(button)

    def set_arrangement_overdub_button(self, button):
        self._arrangement_overdub_toggle.set_toggle_button(button)

    def set_overdub_button(self, button):
        self._overdub_toggle.set_toggle_button(button)

    def set_tempo_control(self, control, fine_control=None):
        if self._tempo_control != control:
            self._tempo_control = control
            self._TransportComponent__tempo_value.subject = control
        if self._tempo_fine_control != fine_control:
            self._tempo_fine_control = fine_control
            self._TransportComponent__tempo_fine_value.subject = fine_control
            self._fine_tempo_needs_pickup = True
            self._prior_fine_tempo_value = -1

    def set_tempo_fine_control(self, fine_control):
        if self._tempo_fine_control != fine_control:
            self._tempo_fine_control = fine_control
            self._TransportComponent__tempo_fine_value.subject = fine_control
            self._fine_tempo_needs_pickup = True
            self._prior_fine_tempo_value = -1

    @listens("value")
    def __ffwd_value_slot(self, value):
        self._ffwd_value(value)

    def _ffwd_value(self, value):
        if self._ffwd_button.is_momentary():
            self._ffwd_task.kill()
            if value:
                self._ffwd_task = self._tasks.add(partial(self._move_current_song_time, SEEK_SPEED))
        else:
            if self.is_enabled():
                self.song.current_song_time += 1

    @listens("value")
    def __rwd_value_slot(self, value):
        self._rwd_value(value)

    def _rwd_value(self, value):
        if self._rwd_button.is_momentary():
            self._rwd_task.kill()
            if value:
                self._rwd_task = self._tasks.add(partial(self._move_current_song_time, -SEEK_SPEED))
        else:
            if self.is_enabled():
                song = self.song
                song.current_song_time = max(0.0, song.current_song_time - 1)

    def _move_current_song_time(self, speed, delta):
        song = self.song
        song.current_song_time = max(0.0, song.current_song_time + speed * delta)
        return task.RUNNING

    @tap_tempo_button.pressed
    def tap_tempo_button(self, _):
        if not self._end_undo_step_task.is_running:
            self.song.begin_undo_step()
        self._end_undo_step_task.restart()
        self.song.tap_tempo()

    @listens("value")
    def __tempo_value(self, value):
        if self.is_enabled():
            fraction = old_div(TEMPO_TOP - TEMPO_BOTTOM, 127.0)
            self.song.tempo = fraction * value + TEMPO_BOTTOM

    @listens("value")
    def __tempo_fine_value(self, value):
        if self.is_enabled():
            if self._fine_tempo_needs_pickup:
                if in_range(self._prior_fine_tempo_value, 0, 128):
                    range_max = max(value, self._prior_fine_tempo_value)
                    range_min = min(value, self._prior_fine_tempo_value)
                    if in_range(64, range_min, range_max + 1):
                        self._fine_tempo_needs_pickup = False
            else:
                difference = value - self._prior_fine_tempo_value
                ratio = old_div(127.0, TEMPO_FINE_RANGE)
                new_tempo = clamp(self.song.tempo + old_div(difference, ratio), TEMPO_BOTTOM, TEMPO_TOP)
                self.song.tempo = new_tempo
        self._prior_fine_tempo_value = value



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_MxDCore/MxDCore.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 67065 bytes
import contextlib
import json
import logging
from functools import partial, reduce, wraps

import _Framework
import Live.Base
from _Framework import Disconnectable
from ableton.v2.base import old_hasattr
from future.utils import string_types

from .LomTypes import (
    CONTROL_SURFACES,
    ENUM_TYPES,
    LIVE_APP,
    PROPERTY_TYPES,
    ROOT_KEYS,
    LomAttributeError,
    LomNoteOperationError,
    LomNoteOperationWarning,
    LomObjectError,
    MFLPropertyFormats,
    data_dict_to_json,
    get_exposed_lom_types,
    get_exposed_property_info,
    get_root_prop,
    is_control_surface,
    is_cplusplus_lom_object,
    is_lom_object,
    is_object_iterable,
    verify_object_property,
)
from .LomUtils import (
    LomInformation,
    LomIntrospection,
    LomPathCalculator,
    LomPathResolver,
    is_control_surfaces_list,
    wrap_control_surfaces_list,
)
from .MxDControlSurfaceAPI import MxDControlSurfaceAPI
from .MxDUtils import StringHandler, TupleWrapper
from .NotesAPIUtils import (
    MIDI_NOTE_ATTRS,
    VALID_DUPLICATE_NOTES_BY_ID_PARAMETERS,
    midi_note_to_dict,
    verify_note_specification_requirements,
)

logger = logging.getLogger(__name__)


def get_current_max_device(device_id):
    return MxDCore.instance.manager.get_max_device(device_id)


def sanitize_list(passed_list, valid_elements):
    passed_elements = set(passed_list)
    sanitized = passed_elements.intersection(valid_elements)
    invalid = passed_elements - sanitized
    return (sanitized, invalid)


PATH_KEY = "CURRENT_PATH"
ID_KEY = "CURRENT_LOM_ID"
TYPE_KEY = "CURRENT_TYPE"
PROP_KEY = "CURRENT_PROPERTY"
PROP_LISTENER_KEY = "PROPERTY_LISTENER"
PATH_LISTENER_KEY = "PATH_LISTENERS"
OPEN_OPERATIONS_KEY = "OPEN_OPERATION"
NOTE_BUFFER_KEY = "NOTE_BUFFER"
NOTE_OPERATION_KEY = "NOTE_OPERATION"
NOTE_COUNT_KEY = "NOTE_COUNT"
NOTE_REPLACE_KEY = "NOTE_REPLACE"
NOTE_SET_KEY = "NOTE_SET"
CONTAINS_CS_ID_KEY = "CONTAINS_CS_ID_KEY"
LAST_SENT_ID_KEY = "LAST_SENT_ID"
INVALID_DICT_ENTRY_ERROR = "Invalid entry in the dictionary"
INVALID_ID_ERROR = "Provide a list of valid note IDs or a dictionary with function parameters as keys"
MALFORMATTED_DICTIONARY_ERROR = "Malformatted dictionary argument"
NOTES_API_MAIN_KEY_ERROR = "Expecting 'notes' as the main dictionary's key"
NOTE_IDS_MISSING_ERROR = "Required key 'note_ids' is missing"
NOTE_ID_MISSING_ERROR = "Required key 'note_id' is missing"
PARSE_ERROR = "Error parsing parameters"
PRIVATE_PROP_WARNING = "Warning: Calling private property. This property might change or be removed in the future."
WARP_MARKER_SPEC_INCOMPLETE_ERROR = "Both beat time and sample time need to be specified"
REGISTER_TIMEABLE_ERROR_NOERROR = "0"
REGISTER_TIMEABLE_ERROR_GENERICERROR = "1"


def concatenate_strings(string_list, string_format="%s %s"):
    return str(reduce(lambda s1, s2: string_format % (s1, s2), string_list) if len(string_list) > 0 else "")


def parameter_to_bool(parameter):
    bool_value = False
    if isinstance(parameter, (int, bool)):
        bool_value = parameter
    else:
        if str(parameter) in ("True", "False"):
            bool_value = str(parameter) == "True"
    return bool_value


def note_from_parameters(parameters):
    new_note = [parameters[0], parameters[1], parameters[2]]
    new_note.append(int(parameters[3]) if (len(parameters) > 3) and (isinstance(parameters[3], (int, float))) else 100)
    new_note.append(len(parameters) > 4 and parameter_to_bool(parameters[4]))
    return tuple(new_note)


def get_object_type_name(obj):
    if is_control_surface(obj):
        return obj.type_name
    return obj.__class__.__name__


class MxDCore:
    instance = None

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self.device_contexts = {}
        self.manager = None
        self.lom_classes = []
        self.epii_version = (-1, -1)
        self._cs_api = MxDControlSurfaceAPI(self)
        self._call_handler = {
            "get_notes": self._object_get_notes_handler,
            "set_notes": self._object_set_notes_handler,
            "get_selected_notes": self._object_selected_notes_handler,
            "replace_selected_notes": self._object_replace_selected_notes_handler,
            "get_all_notes_extended": self._object_get_notes_extended_handler,
            "get_notes_by_id": self._object_get_notes_by_id_handler,
            "get_notes_extended": self._object_get_notes_extended_handler,
            "get_selected_notes_extended": self._object_get_notes_extended_handler,
            "add_new_notes": self._object_add_new_notes_handler,
            "apply_note_modifications": self._object_apply_note_modifications_handler,
            "duplicate_notes_by_id": self._object_duplicate_notes_by_id_handler,
            "remove_notes_by_id": self._object_perform_operation_on_notes_by_id_handler,
            "select_notes_by_id": self._object_perform_operation_on_notes_by_id_handler,
            "notes": self._object_notes_handler,
            "note": self._object_note_handler,
            "done": self._object_done_handler,
            "get_control_names": self._cs_api.object_get_control_names,
            "get_control": self._cs_api.object_get_control,
            "grab_control": self._cs_api.object_grab_control,
            "release_control": self._cs_api.object_release_control,
            "send_midi": self._cs_api.object_send_midi,
            "send_receive_sysex": self._cs_api.object_send_receive_sysex,
            "grab_midi": self._cs_api.object_grab_midi,
            "release_midi": self._cs_api.object_release_midi,
            "add_warp_marker": self._object_warp_marker_handler,
        }
        self.lom_classes = get_exposed_lom_types()
        self.lom_classes += LomIntrospection(_Framework).lom_classes
        self.appointed_lom_ids = {0: None}

    def disconnect(self):
        for dev_id in list(self.device_contexts.keys()):
            device_context = self.device_contexts[dev_id]
            if device_context[CONTAINS_CS_ID_KEY]:
                self.release_device_context(dev_id)

        TupleWrapper.forget_tuple_wrapper_instances()
        self.manager.set_manager_callbacks(None, None, None, None)
        self.manager = None
        del self.appointed_lom_ids
        del self.lom_classes

    def set_manager(self, manager):
        self.manager = manager
        manager.set_manager_callbacks(
            self.update_observer_listener,
            self.install_observer_listener,
            self.uninstall_observer_listener,
            self.update_timeable,
        )
        self.epii_version = manager.get_epii_version()

    def _get_lom_object_by_lom_id(self, referring_device_id, lom_id):
        if lom_id > 0:
            return self.manager.get_lom_object(referring_device_id, lom_id)
        return self.appointed_lom_ids[lom_id]

    def _lom_id_exists(self, referring_device_id, lom_id):
        if lom_id > 0:
            return self.manager.lom_id_exists(referring_device_id, lom_id)
        return lom_id in self.appointed_lom_ids

    def _get_lom_id_by_lom_object(self, lom_object):
        if is_cplusplus_lom_object(lom_object):
            return self.manager.get_lom_id(lom_object)
        for id, object in self.appointed_lom_ids.items():
            if object == lom_object:
                return id

        id = -len(self.appointed_lom_ids)
        self.appointed_lom_ids[id] = lom_object
        if isinstance(lom_object, Disconnectable):

            def unregister_lom_object(f):
                @wraps(f)
                def wrapper(*a, **k):
                    with contextlib.suppress(KeyError):
                        del self.appointed_lom_ids[id]

                    return f(*a, **k)

                return wrapper

            lom_object.disconnect = unregister_lom_object(lom_object.disconnect)
        return id

    def _get_lom_ids_of_mapped_objects(self, lom_id):
        observers = []
        remotes = []
        modulators = []
        for device_id, device_context in self.device_contexts.items():
            for object_id, _ in device_context.items():
                if not isinstance(object_id, int):
                    continue
                else:
                    type = self._get_current_type(device_id, object_id)
                if type == "obs":
                    if self._get_current_lom_id(device_id, object_id) == int(lom_id):
                        observers.append((device_id, object_id))
                else:
                    if type == "rmt":
                        if self._get_current_lom_id(device_id, object_id) == int(lom_id):
                            remotes.append((device_id, object_id))
                    else:
                        if type == "mod":
                            if self._get_current_lom_id(device_id, object_id) == int(lom_id):
                                modulators.append((device_id, object_id))

        return (observers, remotes, modulators)

    def _get_object_path(self, device_id, lom_object):
        resolver = LomPathCalculator(lom_object, get_current_max_device(device_id))
        return concatenate_strings(resolver.path_components)

    def _is_integer(self, s):
        if s[0] in ("-", "+"):
            return s[1:].isdigit()
        return s.isdigit()

    def _set_current_lom_id(self, device_id, object_id, lom_id, type):
        device_context = self.device_contexts[device_id]
        if self.manager.set_current_lom_id(device_id, object_id, lom_id):
            device_context[object_id][ID_KEY] = 0
        else:
            device_context[object_id][ID_KEY] = lom_id
            self._set_current_type(device_id, object_id, type)
            if type == "obs":
                self._observer_update_listener(device_id, object_id)
            else:
                if type == "rmt":
                    self._update_timeable(device_id, object_id, 4, True)
                else:
                    if type == "mod":
                        self._update_timeable(device_id, object_id, 5, True)
        device_context[CONTAINS_CS_ID_KEY] |= lom_id in self.appointed_lom_ids

    def _get_current_lom_id(self, device_id, object_id):
        current_id = self.manager.get_current_lom_id(device_id, object_id)
        if current_id != 0:
            return current_id
        return self.device_contexts[device_id][object_id][ID_KEY]

    def _set_current_lom_id_from_param(self, device_id, object_id, object_type, parameter):
        if self._is_integer(parameter) and self._lom_id_exists(device_id, int(parameter)):
            new_id = int(parameter)
            lom_object = self._get_lom_object_by_lom_id(device_id, new_id)
            if isinstance(lom_object, (Live.DeviceParameter.DeviceParameter, type(None))):
                self._set_current_lom_id(device_id, object_id, new_id, object_type)
            else:
                self._print_error(device_id, object_id, "set id: only accepts objects of type DeviceParameter")
        else:
            self._print_error(device_id, object_id, "set id: invalid id")

    def _set_current_type(self, device_id, object_id, type):
        if type == "obj":
            old_type = self.device_contexts[device_id][object_id][TYPE_KEY]
            if old_type is None:
                self.device_contexts[device_id][object_id][TYPE_KEY] = type
        else:
            self.device_contexts[device_id][object_id][TYPE_KEY] = type

    def _get_current_type(self, device_id, object_id):
        current_type = self.manager.get_type(device_id, object_id)
        if current_type != -1:
            return {0: "obs", 1: None, 2: "obj", 3: "obs", 4: "rmt", 5: "mod"}[current_type]
        return self.device_contexts[device_id][object_id][TYPE_KEY]

    def _set_current_property(self, device_id, object_id, property_name):
        if not self.manager.set_current_property(device_id, object_id, property_name):
            self.device_contexts[device_id][object_id][PROP_KEY] = property_name
            self._set_current_type(device_id, object_id, "obs")
            self._observer_update_listener(device_id, object_id)

    def _get_current_property(self, device_id, object_id):
        property_name = self.manager.get_current_property(device_id, object_id)
        if property_name != "":
            return property_name
        return self.device_contexts[device_id][object_id][PROP_KEY]

    def update_device_context(self, device_id, object_id):
        if device_id not in list(self.device_contexts.keys()):
            self.device_contexts[device_id] = {CONTAINS_CS_ID_KEY: False}
        if object_id not in list(self.device_contexts[device_id].keys()):
            self.device_contexts[device_id][object_id] = {
                PATH_KEY: [],
                ID_KEY: 0,
                TYPE_KEY: None,
                PROP_KEY: "",
                PROP_LISTENER_KEY: (None, None, None),
                PATH_LISTENER_KEY: {},
                OPEN_OPERATIONS_KEY: {},
                LAST_SENT_ID_KEY: None,
            }

    def release_device_context(self, device_id, *_):
        device_context = self.device_contexts[device_id]
        for key in list(device_context.keys()):
            if isinstance(key, int):
                object_context = device_context[key]
                self._observer_uninstall_listener(device_id, key)
                self._cs_api.release_control_surface_midi(device_id, key)
                if len(object_context[PATH_KEY]) > 0:
                    object_context[PATH_KEY] = []
                    self._install_path_listeners(device_id, key, self._path_listener_callback)

        del self.device_contexts[device_id]

    def prepare_control_surface_update(self, *_):
        found_cs_references = False
        for device_id in list(self.device_contexts.keys()):
            device_context = self.device_contexts[device_id]
            if device_context[CONTAINS_CS_ID_KEY]:
                found_cs_references = True
                self.release_device_context(device_id)
                self.manager.refresh_max_device(device_id)

        if found_cs_references:
            TupleWrapper.forget_tuple_wrapper_instances()
            self.appointed_lom_ids = {0: None}
        self._cs_api.wrapper_registry.clear()

    def path_set_path(self, device_id, object_id, parameters):
        pure_path = parameters.strip().strip('"')
        path_components = pure_path.split(" ")
        if len(pure_path) > 0 and path_components[0] not in ROOT_KEYS:
            self._print_error(device_id, object_id, "set path: invalid path")
        else:
            self.device_contexts[device_id][object_id][PATH_KEY] = []
            self.path_goto(device_id, object_id, parameters)

    def path_goto(self, device_id, object_id, parameters):
        self._goto_path(device_id, object_id, parameters)
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        result_object = self._object_from_path(device_id, object_id, (object_context[PATH_KEY]), must_exist=False)
        result_id = str(self._get_lom_id_by_lom_object(result_object))
        device_context[CONTAINS_CS_ID_KEY] |= CONTROL_SURFACES in object_context[PATH_KEY]
        result_path = str(concatenate_strings(object_context[PATH_KEY]))
        for msg_type, value in (
            ("path_curr_path", result_path),
            ("path_orig_id", result_id),
            ("path_curr_id", result_id),
        ):
            self.manager.send_message(device_id, object_id, msg_type, value)

        self._install_path_listeners(device_id, object_id, self._path_listener_callback)

    def path_get_id(self, device_id, object_id, parameters):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        lom_object = self._object_from_path(device_id, object_id, (object_context[PATH_KEY]), must_exist=False)
        result_id = str(self._get_lom_id_by_lom_object(lom_object))
        for msg_type, value in (("path_orig_id", result_id), ("path_curr_id", result_id)):
            self.manager.send_message(device_id, object_id, msg_type, value)

    def path_bang(self, device_id, object_id, parameters):
        self.path_get_id(device_id, object_id, parameters)

    def _get_path_and_object(self, device_id, object_id):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        current_path = object_context[PATH_KEY]
        current_object = self._object_from_path(device_id, object_id, current_path, must_exist=True)
        return (current_path, current_object)

    def _get_lom_object_properties(self, device_id, object_id, looking_for):
        current_path, current_object = self._get_path_and_object(device_id, object_id)
        if len(current_path) == 0:
            result = concatenate_strings(ROOT_KEYS)
        else:
            if current_object is not None:
                current_object = self._disambiguate_object(current_object)
                if is_object_iterable(current_object):
                    result = "%d list elements, no %s" % (len(current_object), looking_for)
                else:
                    lom_info = LomInformation(current_object, self.epii_version)
                    path_props = [info[0] for info in lom_info.lists_of_children + lom_info.children]
                    result = concatenate_strings(sorted(path_props))
        return result

    def path_get_props(self, device_id, object_id, parameters):
        result = self._get_lom_object_properties(device_id, object_id, "properties") or "No path properties"
        self.manager.send_message(device_id, object_id, "path_props", result)

    def path_get_children(self, device_id, object_id, parameters):
        result = self._get_lom_object_properties(device_id, object_id, "children") or "No children"
        self.manager.send_message(device_id, object_id, "path_children", result)

    def path_get_count(self, device_id, object_id, parameters):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        current_path = object_context[PATH_KEY]
        current_object = self._object_from_path(device_id, object_id, current_path, must_exist=True)
        property = None
        if len(current_path) == 0:
            if parameters in ROOT_KEYS:
                property = get_root_prop(get_current_max_device(device_id), parameters)
        else:
            if current_object is not None:
                if old_hasattr(current_object, parameters):
                    property = getattr(current_object, parameters)
        if property is not None:
            count = str(len(property) if is_object_iterable(property) else -1)
            self.manager.send_message(device_id, object_id, "path_count", concatenate_strings((parameters, count)))
        else:
            self._print_error(device_id, object_id, "getcount: invalid property name")

    def obj_set_id(self, device_id, object_id, parameter):
        if self._is_integer(parameter) and self._lom_id_exists(device_id, int(parameter)):
            self._set_current_lom_id(device_id, object_id, int(parameter), "obj")
        else:
            self._print_error(device_id, object_id, "set id: invalid id")

    def obj_get_id(self, device_id, object_id, parameter):
        self.manager.send_message(device_id, object_id, "obj_id", str(self._get_current_lom_id(device_id, object_id)))

    def obj_get_path(self, device_id, object_id, parameters):
        lom_object = self._get_current_lom_object(device_id, object_id)
        path = self._get_object_path(device_id, lom_object)
        if len(path) == 0 and lom_object is not None:
            self._print_error(device_id, object_id, "get path: error calculating the path")
        else:
            self.manager.send_message(device_id, object_id, "obj_path", str(path).strip())

    def obj_get_type(self, device_id, object_id, parameters):
        current_object = self._get_current_lom_object(device_id, object_id)
        object_type = "unknown"
        if current_object is not None:
            current_object = self._disambiguate_object(current_object)
            object_type = get_object_type_name(current_object)
        self.manager.send_message(device_id, object_id, "obj_type", str(object_type))

    def obj_get_info(self, device_id, object_id, parameters):
        current_object = self._get_current_lom_object(device_id, object_id)
        object_info = "No object"
        if current_object is not None:
            object_info = f"id {self._get_lom_id_by_lom_object(current_object)!s}\n"
            current_object = self._disambiguate_object(current_object)
            lom_info = LomInformation(current_object, self.epii_version)
            object_info += f"type {get_object_type_name(current_object)!s}\n"
            object_info += f"{lom_info.description}\n"
            if not is_object_iterable(current_object):

                def accumulate_info(info_list, label):
                    result = ""
                    if len(info_list) > 0:
                        str_format = "%s %s %s\n" if len(info_list[0]) > 1 else "%s %s\n"
                        formatter = lambda info: str_format % ((label, *info))
                        formatted_list = list(map(formatter, info_list))
                        sorted_list = sorted(formatted_list)
                        result = concatenate_strings(sorted_list, string_format="%s%s")
                    return result

                object_info += (
                    accumulate_info(lom_info.lists_of_children, "children")
                    + accumulate_info(lom_info.children, "child")
                    + accumulate_info(lom_info.properties, "property")
                    + accumulate_info(lom_info.functions, "function")
                )
            object_info += "done"
        self.manager.send_message(device_id, object_id, "obj_info", str(object_info))

    def obj_set_val(self, device_id, object_id, parameters):
        self.obj_set(device_id, object_id, parameters)

    def _set_property_value(self, lom_object, property_name, value):
        verify_object_property(lom_object, property_name, self.epii_version)
        prop = getattr(lom_object, property_name)
        prop_info = get_exposed_property_info(type(lom_object), property_name, self.epii_version)
        if prop_info and prop_info.from_json:
            value = prop_info.from_json(lom_object, value)
            if value is None:
                raise LomAttributeError("set: invalid value")
        else:
            if property_name in list(PROPERTY_TYPES.keys()):
                if not is_lom_object(value, self.lom_classes):
                    raise LomAttributeError("set: no valid object id")
                if not isinstance(value, PROPERTY_TYPES[property_name]):
                    raise LomAttributeError("set: type mismatch")
            else:
                if isinstance(prop, (int, bool)):
                    if str(value) in ("True", "False"):
                        value = int(str(value) == "True")
                    else:
                        if not isinstance(value, int):
                            raise LomAttributeError("set: invalid value")
                else:
                    if isinstance(prop, float):
                        if not isinstance(value, (int, float)):
                            raise LomAttributeError("set: type mismatch")
                        value = float(value)
                    else:
                        if isinstance(prop, string_types):
                            if not isinstance(value, (string_types, int, float)):
                                raise LomAttributeError("set: type mismatch")
                            value = str(value)
                        else:
                            if isinstance(prop, tuple):
                                prop_info = get_exposed_property_info(
                                    type(lom_object),
                                    property_name,
                                    self.epii_version,
                                )
                                if prop_info and prop_info.format == MFLPropertyFormats.JSON:
                                    package = json.loads(value)
                                    value = tuple(package[property_name])
                                else:
                                    raise LomAttributeError("set: unsupported property type")
                            else:
                                raise LomAttributeError("set: unsupported property type")
        setattr(lom_object, property_name, value)

    def _warn_if_using_private_property(self, device_id, object_id, property_name):
        did_warn = False
        warn = partial(self._print_warning, device_id, object_id)
        if property_name.startswith("_"):
            warn(PRIVATE_PROP_WARNING)
            did_warn = True
        return did_warn

    def obj_set(self, device_id, object_id, parameters):
        current_object = self._get_current_lom_object(device_id, object_id)
        if current_object is not None:
            parsed_params = self._parse(device_id, object_id, parameters)
            property_name = parsed_params[0]
            property_values = parsed_params[1:]
            value = property_values[0]
            try:
                self._set_property_value(current_object, property_name, value)
                self._warn_if_using_private_property(device_id, object_id, property_name)
            except LomAttributeError as e:
                try:
                    self._print_error(device_id, object_id, str(e))
                finally:
                    e = None
                    del e

        else:
            self._print_error(device_id, object_id, "set: no valid object set")

    def obj_get_val(self, device_id, object_id, parameters):
        self.obj_get(device_id, object_id, parameters)

    def obj_get(self, device_id, object_id, parameters):
        current_object = self._get_current_lom_object(device_id, object_id)
        result_value = None
        if current_object is not None:
            try:
                if parameters.isdigit():
                    result_value = current_object[int(parameters)]
                else:
                    if not self._warn_if_using_private_property(device_id, object_id, parameters):
                        verify_object_property(current_object, parameters, self.epii_version)
                    result_value = self._get_lom_object_prop(current_object, parameters)
                    if isinstance(result_value, ENUM_TYPES):
                        result_value = int(result_value)
                prop_info = get_exposed_property_info(type(current_object), parameters, self.epii_version)
                if prop_info and prop_info.to_json:
                    result = self.str_representation_for_object(prop_info.to_json(current_object))
                else:
                    result = self.str_representation_for_object(result_value)
                self.manager.send_message(device_id, object_id, "obj_prop_val", result)
            except LomAttributeError as e:
                try:
                    self._print_error(device_id, object_id, str(e))
                finally:
                    e = None
                    del e

        else:
            self._print_error(device_id, object_id, "get: no valid object set")

    def obj_call(self, device_id, object_id, parameters):
        current_object = self._get_current_lom_object(device_id, object_id)
        param_comps = None
        try:
            param_comps = self._parse(device_id, object_id, parameters)
        except Exception:
            self._print_error(device_id, object_id, f"{PARSE_ERROR}: {parameters}")
            return
        else:

            def to_str(param_comps):
                return " ".join(map(str, param_comps))

        if current_object is not None:
            try:
                func_name = str(param_comps[0])
                handler = (
                    self._call_handler[func_name]
                    if func_name in list(self._call_handler.keys())
                    else self._object_default_call_handler
                )
                handler(device_id, object_id, current_object, param_comps)
            except AttributeError as e:
                try:
                    self._print_error(device_id, object_id, str(e))
                finally:
                    e = None
                    del e

            except RuntimeError as e:
                try:
                    self._print_error(device_id, object_id, f"{e!s}: '{to_str(param_comps)}'")
                finally:
                    e = None
                    del e

            except Exception as e:
                try:
                    reason = "Invalid " + ("arguments" if isinstance(e, TypeError) else "syntax")
                    self._print_error(device_id, object_id, f"{reason}: '{to_str(param_comps)}'")
                finally:
                    e = None
                    del e

        else:
            self._print_error(device_id, object_id, f"call {to_str(param_comps)}: no valid object set")

    def obs_set_id(self, device_id, object_id, parameter):
        if self._is_integer(parameter) and self._lom_id_exists(device_id, int(parameter)):
            self.device_contexts[device_id][object_id][LAST_SENT_ID_KEY] = None
            self._set_current_lom_id(device_id, object_id, int(parameter), "obs")
        else:
            self._print_error(device_id, object_id, "set id: invalid id")

    def obs_get_id(self, device_id, object_id, parameter):
        self.manager.send_message(device_id, object_id, "obs_id", str(self._get_current_lom_id(device_id, object_id)))

    def obs_set_prop(self, device_id, object_id, parameter):
        self._set_current_property(device_id, object_id, parameter)

    def obs_get_prop(self, device_id, object_id, parameter):
        self.manager.send_message(
            device_id,
            object_id,
            "obs_prop",
            str(self._get_current_property(device_id, object_id)),
        )

    def obs_get_type(self, device_id, object_id, parameter):
        current_object = self._get_current_lom_object(device_id, object_id)
        property_name = self._get_current_property(device_id, object_id)
        result = "unknown"
        if current_object is not None:
            if property_name != "":
                if old_hasattr(current_object, property_name):
                    result = getattr(current_object, property_name).__class__.__name__
                else:
                    self._print_warning(device_id, object_id, "gettype: no property with the name " + property_name)
        self.manager.send_message(device_id, object_id, "obs_type", str(result))

    def obs_bang(self, device_id, object_id, parameter):
        self._observer_property_callback(device_id, object_id)

    def rmt_set_id(self, device_id, object_id, parameter):
        self._set_current_lom_id_from_param(device_id, object_id, "rmt", parameter)

    def mod_set_id(self, device_id, object_id, parameter):
        self._set_current_lom_id_from_param(device_id, object_id, "mod", parameter)

    def _object_attr_path_iter(self, device_id, object_id, path_components):
        if len(path_components) == 0:
            return
        cur_object = get_root_prop(get_current_max_device(device_id), path_components[0])
        for component in path_components[1:]:
            if cur_object is None:
                return
            yield (cur_object, component)
            if component.isdigit():
                index = int(component)
                if index >= 0 and index < len(cur_object):
                    cur_object = cur_object[index]
                else:
                    return
            else:
                try:
                    cur_object = getattr(cur_object, component)
                except Exception:
                    return

    def _object_from_path(self, device_id, object_id, path_components, must_exist):
        lom_object = None
        try:
            resolver = LomPathResolver(
                path_components,
                (get_current_max_device(device_id)),
                (self.lom_classes),
                (self.manager),
                cs_wrapper_registry=(self._cs_api.wrapper_registry),
            )
            lom_object = resolver.lom_object
        except (LomAttributeError, LomObjectError) as e:
            try:
                if must_exist or isinstance(e, LomObjectError):
                    self._print_error(device_id, object_id, str(e))
            finally:
                e = None
                del e

        return lom_object

    def _get_current_lom_object(self, device_id, object_id):
        return self._get_lom_object_by_lom_id(device_id, self._get_current_lom_id(device_id, object_id))

    def _object_for_id(self, device_id):
        with contextlib.suppress(Exception):
            return partial(self._get_lom_object_by_lom_id, device_id)

    def str_representation_for_object(self, lom_object, mark_ids=True):
        result = ""
        lom_object = self._disambiguate_object(lom_object)
        if is_object_iterable(lom_object):
            result = concatenate_strings(list(map(self.str_representation_for_object, lom_object)))
        else:
            if is_lom_object(lom_object, self.lom_classes):
                result = ("id " if mark_ids else "") + str(self._get_lom_id_by_lom_object(lom_object))
            else:
                if isinstance(lom_object, (int, bool)):
                    result = str(int(lom_object))
                else:
                    result = StringHandler.prepare_outgoing(str(lom_object))
        return result

    def _install_path_listeners(self, device_id, object_id, listener_callback):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        path_components = object_context[PATH_KEY]
        new_listeners = {}
        self._uninstall_path_listeners(device_id, object_id)
        listener = partial(listener_callback, device_id, object_id)
        obj_attr_iter = self._object_attr_path_iter(device_id, object_id, path_components)
        for lom_object, attribute in obj_attr_iter:
            attribute = self._listenable_property_for(attribute)
            if lom_object is not None:
                if old_hasattr(lom_object, attribute + "_has_listener"):
                    getattr(lom_object, f"add_{attribute}_listener")(listener)
                    new_listeners[(lom_object, attribute)] = listener

        object_context[PATH_LISTENER_KEY] = new_listeners

    def _uninstall_path_listeners(self, device_id, object_id):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        old_listeners = object_context[PATH_LISTENER_KEY]
        for (lom_object, attribute), listener in old_listeners.items():
            if lom_object is not None:
                if getattr(lom_object, attribute + "_has_listener")(listener):
                    getattr(lom_object, f"remove_{attribute}_listener")(listener)

    def _path_listener_callback(self, device_id, object_id):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        resulting_id = self._get_lom_id_by_lom_object(
            self._object_from_path(device_id, object_id, (object_context[PATH_KEY]), must_exist=False),
        )
        self.manager.send_message(device_id, object_id, "path_curr_id", str(resulting_id))
        self._install_path_listeners(device_id, object_id, self._path_listener_callback)

    def _goto_path(self, device_id, object_id, parameters):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        try:
            pure_path = parameters.strip().strip('"')
            path_components = []
            if len(pure_path) > 0:
                path_components = pure_path.strip().split(" ")
            if tuple(path_components[:2]) == (LIVE_APP, CONTROL_SURFACES):
                object_context[PATH_KEY] = [LIVE_APP, CONTROL_SURFACES]
                path_components = path_components[2:]
            for parameter in path_components:
                if parameter == "up":
                    del object_context[PATH_KEY][-1]
                else:
                    if parameter in ROOT_KEYS:
                        object_context[PATH_KEY] = []
                    object_context[PATH_KEY].append(parameter)

        except:
            self._print_error(device_id, object_id, "goto: invalid path")
            return

    def _object_default_call_handler(self, device_id, object_id, lom_object, parameters):
        verify_object_property(lom_object, parameters[0], self.epii_version)
        self._warn_if_using_private_property(device_id, object_id, parameters[0])
        function = getattr(lom_object, parameters[0])
        result = function(*parameters[1:])
        result_str = self.str_representation_for_object(result)
        self.manager.send_message(device_id, object_id, "obj_call_result", result_str)

    def _object_get_notes_handler(self, device_id, object_id, lom_object, parameters):
        notes = lom_object.get_notes(parameters[1], parameters[2], parameters[3], parameters[4])
        self.manager.send_message(device_id, object_id, "obj_call_result", self._create_notes_output(notes))

    def _extract_dict_from_parameters(self, parameters):
        if not len(parameters) > 0 or isinstance(parameters[0], string_types):
            with contextlib.suppress(ValueError):
                return json.loads(parameters[0])
        return None

    def _object_get_notes_by_id_handler(self, device_id, object_id, lom_object, parameters):
        function_name, function_parameters = parameters[0], parameters[1:]
        dict_from_args = self._extract_dict_from_parameters(function_parameters)
        self._do_get_notes_extended(
            device_id,
            object_id,
            lom_object,
            function_name,
            dict_from_args,
            function_parameters if (not dict_from_args) else ([]),
        )

    def _object_get_notes_extended_handler(self, device_id, object_id, lom_object, parameters):
        function_name, function_parameters = parameters[0], parameters[1:]
        dict_from_args = self._extract_dict_from_parameters(function_parameters)
        function_parameters = function_parameters if (not dict_from_args) else ([])
        (self._do_get_notes_extended)(
            device_id,
            object_id,
            lom_object,
            function_name,
            dict_from_args,
            *function_parameters,
        )

    def _do_get_notes_extended(self, device_id, object_id, lom_object, function_name, param_dict, *function_parameters):
        verify_object_property(lom_object, function_name, self.epii_version)
        properties_to_return = None
        if param_dict is not None:
            properties_to_return = param_dict.pop("return", None)
            if properties_to_return is not None:
                properties_to_return = self._sanitize_midi_note_property_list(
                    device_id,
                    object_id,
                    properties_to_return,
                )
        try:
            func = getattr(lom_object, function_name)
            midi_note_vector = func(**param_dict) if param_dict is not None else func(*function_parameters)
        except ValueError as e:
            try:
                raise RuntimeError(str(e))
            finally:
                e = None
                del e

        result = self.str_representation_for_object(
            self._midi_note_vector_to_dict_output(midi_note_vector, properties_to_return),
        )
        self.manager.send_message(device_id, object_id, "obj_call_result", result)

    def _object_add_new_notes_handler(self, device_id, object_id, lom_object, parameters):
        function_name, function_parameter = parameters[0], parameters[1]
        verify_object_property(lom_object, function_name, self.epii_version)
        note_dicts = self._get_list_of_note_dictionaries(function_parameter)
        note_specifications = []
        for note_dict in note_dicts:
            verify_note_specification_requirements(note_dict)
            try:
                note_specifications.append((Live.Clip.MidiNoteSpecification)(**note_dict))
            except TypeError:
                raise RuntimeError(INVALID_DICT_ENTRY_ERROR)

        result = getattr(lom_object, function_name)(note_specifications)
        self.manager.send_message(device_id, object_id, "obj_call_result", self.str_representation_for_object(result))

    def _object_warp_marker_handler(self, device_id, object_id, lom_object, parameters):
        function_name, function_param = parameters[0], parameters[1]
        warp_marker_dict = json.loads(function_param)
        verify_object_property(lom_object, function_name, self.epii_version)
        dict_keys = warp_marker_dict.keys()
        if "beat_time" not in dict_keys:
            raise RuntimeError(WARP_MARKER_SPEC_INCOMPLETE_ERROR)
        if "sample_time" not in dict_keys:
            num_samples = lom_object.beat_to_sample_time(warp_marker_dict["beat_time"])
            sample_rate = lom_object.sample_rate
            warp_marker_dict["sample_time"] = num_samples / float(sample_rate)
        warp_marker = None
        try:
            warp_marker = (Live.Clip.WarpMarker)(**warp_marker_dict)
        except TypeError:
            raise RuntimeError(INVALID_DICT_ENTRY_ERROR)

        result = getattr(lom_object, function_name)(warp_marker)
        self.manager.send_message(device_id, object_id, "obj_call_result", self.str_representation_for_object(result))

    def _object_apply_note_modifications_handler(self, device_id, object_id, lom_object, parameters):
        function_name, function_parameter = parameters[0], parameters[1]
        verify_object_property(lom_object, function_name, self.epii_version)
        note_dicts = self._get_list_of_note_dictionaries(function_parameter)
        try:
            id_to_note_mapping = {note["note_id"]: note for note in note_dicts}
        except KeyError:
            raise RuntimeError(NOTE_ID_MISSING_ERROR)

        midi_notes = lom_object.get_notes_by_id(id_to_note_mapping.keys())
        for midi_note in midi_notes:
            note_dict = id_to_note_mapping[midi_note.note_id]
            for property_name, value in note_dict.items():
                if property_name != "note_id":
                    setattr(midi_note, property_name, value)

        result = getattr(lom_object, function_name)(midi_notes)
        self.manager.send_message(device_id, object_id, "obj_call_result", self.str_representation_for_object(result))

    def _object_duplicate_notes_by_id_handler(self, device_id, object_id, lom_object, parameters):
        function_name, function_parameters = parameters[0], parameters[1:]
        verify_object_property(lom_object, function_name, self.epii_version)
        func = getattr(lom_object, function_name)
        param_dict = self._extract_dict_from_parameters(function_parameters)
        if param_dict is not None:
            if "note_ids" not in param_dict:
                raise RuntimeError(NOTE_IDS_MISSING_ERROR)
            sanitized, invalid = sanitize_list(param_dict.keys(), VALID_DUPLICATE_NOTES_BY_ID_PARAMETERS)
            if invalid:
                for key in invalid:
                    param_dict.pop(key)

                multiple = len(invalid) > 1
                self._print_warning(
                    device_id,
                    object_id,
                    "Invalid key{} provided: {}. {} will be ignored.".format(
                        "s" if multiple else "",
                        ", ".join(invalid),
                        "They" if multiple else "It",
                    ),
                )
            try:
                result = func(**param_dict)
            except ValueError as e:
                try:
                    raise RuntimeError(str(e))
                finally:
                    e = None
                    del e

        else:
            try:
                result = func(function_parameters)
            except TypeError:
                raise RuntimeError(INVALID_ID_ERROR)
            except ValueError as e:
                try:
                    raise RuntimeError(str(e))
                finally:
                    e = None
                    del e

        self.manager.send_message(device_id, object_id, "obj_call_result", self.str_representation_for_object(result))

    def _object_perform_operation_on_notes_by_id_handler(self, device_id, object_id, lom_object, parameters):
        function_name, function_parameters = parameters[0], parameters[1:]
        verify_object_property(lom_object, function_name, self.epii_version)
        try:
            result = getattr(lom_object, function_name)(function_parameters)
        except ValueError as e:
            try:
                raise RuntimeError(str(e))
            finally:
                e = None
                del e

        self.manager.send_message(device_id, object_id, "obj_call_result", self.str_representation_for_object(result))

    def _object_selected_notes_handler(self, device_id, object_id, lom_object, parameters):
        notes = lom_object.get_selected_notes()
        self.manager.send_message(device_id, object_id, "obj_call_result", self._create_notes_output(notes))

    def _object_set_notes_handler(self, device_id, object_id, lom_object, parameters):
        self._start_note_operation(device_id, object_id, lom_object, parameters, NOTE_SET_KEY)

    def _object_replace_selected_notes_handler(self, device_id, object_id, lom_object, parameters):
        self._start_note_operation(device_id, object_id, lom_object, parameters, NOTE_REPLACE_KEY)

    def _object_notes_handler(self, device_id, object_id, lom_object, parameters):
        device_context = self.device_contexts[device_id][object_id]
        if NOTE_OPERATION_KEY in list(device_context[OPEN_OPERATIONS_KEY].keys()):
            device_context[OPEN_OPERATIONS_KEY][NOTE_COUNT_KEY] = parameters[1]
        else:
            self._print_error(device_id, object_id, "no operation in progress")

    def _object_note_handler(self, device_id, object_id, lom_object, parameters):
        device_context = self.device_contexts[device_id][object_id]
        operations = device_context[OPEN_OPERATIONS_KEY]
        try:
            if NOTE_OPERATION_KEY not in operations:
                raise LomNoteOperationError("no operation in progress")
            if NOTE_COUNT_KEY not in operations:
                raise LomNoteOperationError("no note count given")
            if len(operations[NOTE_BUFFER_KEY]) >= operations[NOTE_COUNT_KEY]:
                raise LomNoteOperationError("too many notes")
            operations[NOTE_BUFFER_KEY].append(note_from_parameters(parameters[1:]))
        except LomNoteOperationError as e:
            try:
                self._print_error(device_id, object_id, str(e))
                self._stop_note_operation(device_id, object_id)
            finally:
                e = None
                del e

    def _selector_for_note_operation(self, note_operation):
        if note_operation not in (NOTE_REPLACE_KEY, NOTE_SET_KEY):
            raise LomNoteOperationWarning("invalid note operation")
        if note_operation == NOTE_SET_KEY:
            return "set_notes"
        return "replace_selected_notes"

    def _object_done_handler(self, device_id, object_id, lom_object, parameters):
        device_context = self.device_contexts[device_id][object_id]
        open_operations = device_context[OPEN_OPERATIONS_KEY]
        if NOTE_OPERATION_KEY in open_operations and NOTE_COUNT_KEY in open_operations:
            try:
                notes = tuple(open_operations[NOTE_BUFFER_KEY])
                if len(notes) != open_operations[NOTE_COUNT_KEY]:
                    raise LomNoteOperationWarning("wrong note count")
                operation = open_operations[NOTE_OPERATION_KEY]
                selector = self._selector_for_note_operation(operation)
                getattr(lom_object, selector)(notes)
            except LomNoteOperationWarning as w:
                try:
                    self._print_warning(device_id, object_id, str(w))
                finally:
                    w = None
                    del w

            self._stop_note_operation(device_id, object_id)
        else:
            self._print_error(device_id, object_id, "no operation in progress")

    def _create_notes_output(self, notes):
        element_format = lambda el: str(int(el) if isinstance(el, bool) else el)
        note_format = lambda note: f"note {concatenate_strings(list(map(element_format, note)))}\n"
        return "notes %d\n%sdone" % (
            len(notes),
            concatenate_strings((list(map(note_format, notes))), string_format="%s%s"),
        )

    def _midi_note_vector_to_dict_output(self, notes, properties_to_return):
        return data_dict_to_json("notes", [midi_note_to_dict(note, properties_to_return) for note in notes])

    def _sanitize_midi_note_property_list(self, device_id, object_id, property_list):
        if not isinstance(property_list, list):
            self._print_warning(device_id, object_id, "Note properties must be provided as a list of symbols.")
            return None
        sanitized = []
        invalid = []
        for prop in property_list:
            if prop not in MIDI_NOTE_ATTRS:
                invalid.append(prop)
            else:
                sanitized.append(prop)

        if invalid:
            multiple = len(invalid) > 1
            self._print_warning(
                device_id,
                object_id,
                "Invalid propert{} provided: {}. {} will be ignored.".format(
                    "ies" if multiple else "y",
                    ", ".join(invalid),
                    "They" if multiple else "It",
                ),
            )
        return sanitized

    def _get_list_of_note_dictionaries(self, parameter):
        try:
            data_dict = json.loads(parameter)
        except Exception:
            raise RuntimeError(MALFORMATTED_DICTIONARY_ERROR)

        try:
            note_dicts = data_dict["notes"]
        except KeyError:
            raise RuntimeError(NOTES_API_MAIN_KEY_ERROR)

        return note_dicts

    def _start_note_operation(self, device_id, object_id, lom_object, parameters, operation):
        device_context = self.device_contexts[device_id][object_id]
        if NOTE_OPERATION_KEY not in list(device_context[OPEN_OPERATIONS_KEY].keys()):
            device_context[OPEN_OPERATIONS_KEY][NOTE_BUFFER_KEY] = []
            device_context[OPEN_OPERATIONS_KEY][NOTE_OPERATION_KEY] = operation
        else:
            self._print_error(device_id, object_id, "an operation is already in progress")

    def _stop_note_operation(self, device_id, object_id):
        device_context = self.device_contexts[device_id][object_id]
        for key in (NOTE_OPERATION_KEY, NOTE_BUFFER_KEY, NOTE_COUNT_KEY):
            with contextlib.suppress(KeyError):
                del device_context[OPEN_OPERATIONS_KEY][key]

    def update_observer_listener(self, device_id, object_id):
        self.update_device_context(device_id, object_id)
        self._observer_update_listener(device_id, object_id)

    def install_observer_listener(self, device_id, object_id):
        self.update_device_context(device_id, object_id)
        self._observer_install_listener(device_id, object_id)

    def uninstall_observer_listener(self, device_id, object_id):
        self.update_device_context(device_id, object_id)
        self._observer_uninstall_listener(device_id, object_id)

    def update_lom_ids_of_mapped_objects(self, lom_id):
        observers, remotes, modulators = self._get_lom_ids_of_mapped_objects(lom_id)
        for device_id, object_id in observers:
            self._observer_update_listener(device_id, object_id)

        for device_id, object_id in remotes:
            self._update_timeable(device_id, object_id, 4, False)

        for device_id, object_id in modulators:
            self._update_timeable(device_id, object_id, 5, False)

    def _observer_update_listener(self, device_id, object_id):
        self._observer_uninstall_listener(device_id, object_id)
        self._observer_install_listener(device_id, object_id)

    def _observer_install_listener(self, device_id, object_id):
        current_object = self._get_current_lom_object(device_id, object_id)
        property_name = self._get_current_property(device_id, object_id)
        self._warn_if_using_private_property(device_id, object_id, property_name)
        if property_name == "id":
            self._observer_id_callback(device_id, object_id)
        else:
            if current_object is not None:
                if property_name != "":
                    object_context = self.device_contexts[device_id][object_id]
                    listener_callback = partial(self._observer_property_callback, device_id, object_id)
                    transl_prop_name = self._listenable_property_for(property_name)
                    if old_hasattr(current_object, transl_prop_name + "_has_listener"):
                        getattr(current_object, f"add_{transl_prop_name}_listener")(listener_callback)
                        object_context[PROP_LISTENER_KEY] = (listener_callback, current_object, property_name)
                        listener_callback()
                    else:
                        if old_hasattr(current_object, transl_prop_name):
                            self._print_warning(device_id, object_id, "property cannot be listened to")

    def _observer_uninstall_listener(self, device_id, object_id):
        device_context = self.device_contexts[device_id]
        object_context = device_context[object_id]
        self._uninstall_path_listeners(device_id, object_id)
        listener_callback, current_object, property_name = object_context[PROP_LISTENER_KEY]
        if current_object is not None:
            if listener_callback is not None:
                transl_prop_name = self._listenable_property_for(property_name)
                if getattr(current_object, transl_prop_name + "_has_listener")(listener_callback):
                    getattr(current_object, f"remove_{transl_prop_name}_listener")(listener_callback)
        object_context[PROP_LISTENER_KEY] = (None, None, None)

    def _observer_property_message_type(self, prop, prop_info):
        prop_type = None
        if prop_info and prop_info.format == MFLPropertyFormats.JSON:
            prop_type = "obs_dict_val"
        else:
            if isinstance(prop, string_types):
                prop_type = "obs_string_val"
            else:
                if isinstance(prop, (int, bool)):
                    prop_type = "obs_int_val"
                else:
                    if isinstance(prop, float):
                        prop_type = "obs_float_val"
                    else:
                        if is_object_iterable(prop) or isinstance(prop, TupleWrapper):
                            prop_type = "obs_list_val"
                        else:
                            if is_lom_object(prop, self.lom_classes):
                                prop_type = "obs_id_val"
        return prop_type

    def _observer_property_callback(self, device_id, object_id, *args):
        current_object = self._get_current_lom_object(device_id, object_id)
        property_name = self._get_current_property(device_id, object_id)
        prop_info = get_exposed_property_info(type(current_object), property_name, self.epii_version)
        if len(args) > 0:
            formatter = lambda arg: str(int(arg) if isinstance(arg, bool) else arg)
            args_type = self._observer_property_message_type(args if len(args) > 1 else args[0], prop_info)
            result = concatenate_strings(list(map(formatter, args)))
            self.manager.send_message(device_id, object_id, args_type, result)
        else:
            if current_object is not None:
                if property_name != "":
                    if old_hasattr(current_object, property_name):
                        prop = self._get_lom_object_prop(current_object, property_name)
                        prop_type = self._observer_property_message_type(prop, prop_info)
                        if prop_type is None:
                            self._print_warning(device_id, object_id, "unsupported property type")
                        else:
                            prop_value = self.str_representation_for_object(
                                prop_info.to_json(current_object) if prop_info and (prop_info.to_json) else prop,
                                mark_ids=False,
                            )
                            self.manager.send_message(device_id, object_id, prop_type, prop_value)
                    else:
                        if old_hasattr(current_object, property_name + "_has_listener"):
                            self.manager.send_message(device_id, object_id, "obs_string_val", "bang")
                        else:
                            self._print_warning(device_id, object_id, "property should be listenable")

    def _observer_id_callback(self, device_id, object_id):
        object_context = self.device_contexts[device_id][object_id]
        current_object = self._get_current_lom_object(device_id, object_id)
        self._goto_path(device_id, object_id, self._get_object_path(device_id, current_object))
        self._install_path_listeners(device_id, object_id, self._observer_id_callback)
        current_id = 0 if current_object is None else self._get_current_lom_id(device_id, object_id)
        if current_id != object_context[LAST_SENT_ID_KEY]:
            object_context[LAST_SENT_ID_KEY] = current_id
            self.manager.send_message(device_id, object_id, "obs_id_val", str(current_id))

    def update_timeable(self, device_id, object_id, object_type):
        self.update_device_context(device_id, object_id)
        self._update_timeable(device_id, object_id, object_type, False)

    def reset_all_current_lom_ids(self, device_id):
        if device_id in self.device_contexts:
            device_context = self.device_contexts[device_id]
            for object_id, _ in device_context.items():
                if isinstance(object_id, int):
                    type = self._get_current_type(device_id, object_id)
                    if type in ("obs", "obj", "rmt", "mod"):
                        self._set_current_lom_id(device_id, object_id, 0, type)

    def _update_timeable(self, device_id, object_id, object_type, validate_change_allowed):
        current_object = self._get_current_lom_object(device_id, object_id)
        if isinstance(current_object, Live.DeviceParameter.DeviceParameter):
            success = self.manager.register_timeable(
                device_id,
                object_id,
                object_type,
                current_object,
                validate_change_allowed,
            )
            error_code = REGISTER_TIMEABLE_ERROR_NOERROR if success else REGISTER_TIMEABLE_ERROR_GENERICERROR
            self.manager.send_message(device_id, object_id, "register_timeable_error", error_code)
        else:
            self.manager.unregister_timeable(device_id, object_id, object_type, validate_change_allowed)

    def _disambiguate_object(self, object):
        result = object
        if object.__class__.__name__ in ("ListWrapper", "TupleWrapper"):
            result = object.get_list()
        return result

    def _listenable_property_for(self, prop_name):
        if prop_name == "clip":
            return "has_clip"
        return prop_name

    def _get_lom_object_prop(self, lom_object, property_name):
        if is_control_surfaces_list(property_name):
            return wrap_control_surfaces_list(lom_object, self._cs_api.wrapper_registry)
        return getattr(lom_object, property_name)

    def _parse(self, device_id, object_id, string):
        return StringHandler.parse(string, self._object_for_id(device_id))

    def _print_error(self, device_id, object_id, message):
        logger.error("Error: " + str(message))
        self.manager.print_message(device_id, object_id, "error", str(message))

    def _print_warning(self, device_id, object_id, message):
        logger.warning("Warning: " + str(message))
        self.manager.print_message(device_id, object_id, "warning", str(message))



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/midi.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 884 bytes
from ableton.v2.control_surface.midi import (
    CC_STATUS,
    NOTE_OFF_STATUS,
    NOTE_ON_STATUS,
    PB_STATUS,
    SYSEX_END,
    SYSEX_GENERAL_INFO,
    SYSEX_IDENTITY_REQUEST_ID,
    SYSEX_IDENTITY_REQUEST_MESSAGE,
    SYSEX_IDENTITY_RESPONSE_ID,
    SYSEX_NON_REALTIME,
    SYSEX_START,
    extract_value,
    is_valid_sysex,
)

__all__ = (
    "CC_STATUS",
    "NOTE_OFF_STATUS",
    "NOTE_ON_STATUS",
    "PB_STATUS",
    "SYSEX_END",
    "SYSEX_GENERAL_INFO",
    "SYSEX_IDENTITY_REQUEST_ID",
    "SYSEX_IDENTITY_REQUEST_MESSAGE",
    "SYSEX_IDENTITY_RESPONSE_ID",
    "SYSEX_NON_REALTIME",
    "SYSEX_START",
    "extract_value",
    "is_valid_sysex",
)



from __future__ import annotations

from typing import Any, Callable

import Live.Application
import Live.Base
import Live.Browser
import Live.CcControlDevice
import Live.Chain
import Live.ChainMixerDevice
import Live.Clip
import Live.ClipSlot
import Live.CompressorDevice
import Live.Conversions
import Live.Device
import Live.DeviceIO
import Live.DeviceParameter
import Live.DriftDevice
import Live.DrumChain
import Live.DrumPad
import Live.Eq8Device
import Live.Groove
import Live.GroovePool
import Live.HybridReverbDevice
import Live.Listener
import Live.LomObject
import Live.MaxDevice
import Live.MeldDevice
import Live.MidiMap
import Live.MixerDevice
import Live.PluginDevice
import Live.RackDevice
import Live.RoarDevice
import Live.Sample
import Live.Scene
import Live.ShifterDevice
import Live.SimplerDevice
import Live.Song
import Live.SpectralResonatorDevice
import Live.Track
import Live.WavetableDevice



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control_element.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6871 bytes
import logging
import re
import traceback

from future.utils import iteritems, string_types

from ..base import Disconnectable, Event, EventObject, const, depends, lazy_attribute, nop, second, task
from .resource import StackingResource

logger = logging.getLogger(__name__)


class ControlElementClient:
    def set_control_element(self, control_element, grabbed):
        pass


class ElementOwnershipHandler:
    def handle_ownership_change(self, control, client, status):
        client.set_control_element(control, status)


class OptimizedOwnershipHandler(ElementOwnershipHandler):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._ownership_changes = {}
        self._sequence_number = 0

    def handle_ownership_change(self, control, client, status):
        if (control, client, not status) in self._ownership_changes:
            del self._ownership_changes[(control, client, not status)]
        else:
            self._ownership_changes[(control, client, status)] = self._sequence_number
        self._sequence_number += 1

    @depends(traceback=(const(traceback)))
    def commit_ownership_changes(self, traceback=None):
        notify = super().handle_ownership_change
        while self._ownership_changes:
            notifications = sorted((iteritems(self._ownership_changes)), key=second)
            self._ownership_changes.clear()
            for (control, client, status), _ in notifications:
                try:
                    notify(control, client, status)
                except Exception:
                    logger.error("Error when trying to give control: %s", control.name)
                    traceback.print_exc()

        self._ownership_changes.clear()
        self._sequence_number = 0


_element_list_access_expr = re.compile("(\\w+)\\[([+-]?\\d+)\\]")


@depends(element_container=(const(None)))
def get_element(obj, element_container=None):
    if isinstance(obj, string_types):
        if element_container is None:
            raise RuntimeError("Control elements can only be accessed by name, if an element container is available")
        match = _element_list_access_expr.match(obj)
        if match:
            name = match.group(1)
            index = int(match.group(2))
            obj = getattr(element_container, name)[index]
        else:
            obj = getattr(element_container, obj)
    return obj


class ControlElement(Disconnectable):
    class ProxiedInterface:
        send_midi = nop
        reset_state = nop

        def __init__(self, outer=None, *a, **k):
            (super(ControlElement.ProxiedInterface, self).__init__)(*a, **k)
            self._outer = outer

        @property
        def outer(self):
            return self._outer

    @lazy_attribute
    def proxied_interface(self):
        return self.ProxiedInterface(outer=self)

    canonical_parent = None
    name = ""
    optimized_send_midi = True
    _has_resource = False
    _resource_type = StackingResource
    _has_task_group = False

    @depends(send_midi=None, register_control=None)
    def __init__(
        self,
        name="",
        is_private=False,
        resource_type=None,
        optimized_send_midi=None,
        send_midi=None,
        register_control=None,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self._send_midi = send_midi
        self.name = name
        self.is_private = is_private
        if resource_type is not None:
            self._resource_type = resource_type
        if optimized_send_midi is not None:
            self.optimized_send_midi = optimized_send_midi
        register_control(self)

    def disconnect(self):
        self.reset()
        super().disconnect()

    def send_midi(self, message):
        return self._send_midi(message, optimized=(self.optimized_send_midi))

    def clear_send_cache(self):
        pass

    def reset(self):
        raise NotImplementedError

    def reset_state(self):
        pass

    @property
    def resource(self):
        return self._resource

    @lazy_attribute
    def _resource(self):
        self._has_resource = True
        return self._resource_type(self._on_resource_received, self._on_resource_lost)

    @lazy_attribute
    @depends(parent_task_group=(task.TaskGroup))
    def _tasks(self, parent_task_group=None):
        tasks = parent_task_group.add(task.TaskGroup())
        self._has_task_group = True
        return tasks

    def _on_resource_received(self, client, *a, **k):
        self.notify_ownership_change(client, True)

    def _on_resource_lost(self, client):
        self.notify_ownership_change(client, False)

    @depends(element_ownership_handler=(const(ElementOwnershipHandler())))
    def notify_ownership_change(self, client, grabbed, element_ownership_handler=None):
        element_ownership_handler.handle_ownership_change(self, client, grabbed)


class NotifyingControlElement(EventObject, ControlElement):
    __events__ = (
        Event(
            name="value",
            doc=" Called when the control element receives a MIDI value\n                             from the hardware ",
        ),
    )



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/controls/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1536 bytes
from ableton.v2.control_surface.control import (
    ButtonControlBase,
    Connectable,
    Control,
    ControlManager,
    EncoderControl,
    InputControl,
    MappedControl,
    PlayableControl,
    RadioButtonGroup,
    SendValueControl,
    SendValueMixin,
    StepEncoderControl,
    control_color,
    control_event,
    control_matrix,
    is_internal_parameter,
)

from ..display import Renderable
from .button import ButtonControl, TouchControl
from .control import SendValueEncoderControl, SendValueInputControl
from .control_list import FixedRadioButtonGroup, control_list
from .mapped import MappableButton, MappedButtonControl, MappedSensitivitySettingControl
from .toggle_button import ToggleButtonControl

Renderable.control_base_type = Control
__all__ = (
    "ButtonControl",
    "ButtonControlBase",
    "Connectable",
    "Control",
    "ControlManager",
    "EncoderControl",
    "FixedRadioButtonGroup",
    "InputControl",
    "MappableButton",
    "MappedButtonControl",
    "MappedControl",
    "MappedSensitivitySettingControl",
    "PlayableControl",
    "RadioButtonGroup",
    "SendValueControl",
    "SendValueEncoderControl",
    "SendValueInputControl",
    "SendValueMixin",
    "StepEncoderControl",
    "ToggleButtonControl",
    "TouchControl",
    "control_color",
    "control_event",
    "control_list",
    "control_matrix",
    "is_internal_parameter",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/Task.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 12201 bytes
import functools
import traceback

from past.utils import old_div

from .Dependency import depends
from .Util import const, find_if, nop, print_message, remove_if
from .Util import linear as linear_fn


class TaskError(Exception):
    pass


KILLED = 0
RUNNING = 1
PAUSED = 2


class Task:
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._state = RUNNING
        self._next = []
        self._task_manager = None

    def clear(self):
        pass

    def do_update(self, timer):
        pass

    def do_restart(self):
        pass

    def add_next(self, task):
        self._next.append(task)
        return task

    def update(self, timer):
        if self._state == RUNNING:
            self.do_update(timer)
        return self._state

    def pause(self):
        if self._state != KILLED:
            self._state = PAUSED
        return self

    def resume(self):
        if self._state != KILLED:
            self._state = RUNNING
        return self

    def toggle_pause(self):
        if self._state != KILLED:
            self._state = RUNNING if self._state == PAUSED else PAUSED
        return self

    def restart(self):
        self.do_restart()
        self._state = RUNNING
        if self._task_manager:
            if self._task_manager.find(self) is None:
                manager = self._task_manager
                self._task_manager = None
                manager.add(self)
        return self

    def kill(self):
        self._state = KILLED
        if self._task_manager:
            for task in self._next:
                self._task_manager.add(task)

        return self

    @property
    def is_running(self):
        return self._state == RUNNING

    @property
    def is_paused(self):
        return self._state == PAUSED

    @property
    def is_killed(self):
        return self._state == KILLED

    @property
    def state(self):
        return self._state

    @property
    def parent_task(self):
        return self._task_manager

    def _set_parent(self, manager):
        if self._task_manager:
            if manager:
                raise TaskError("Already attached to: " + str(self._task_manager))
        self._task_manager = manager

    def _task_equivalent(self, other):
        return self == other


class WrapperTask(Task):
    def __init__(self, wrapped_task=None, *a, **k):
        (super().__init__)(*a, **k)
        self.wrapped_task = wrapped_task

    def update(self, delta):
        super().update(delta)
        self.wrapped_task.update(delta)

    def _set_parent(self, manager):
        super()._set_parent(manager)
        self.wrapped_task._set_parent(manager)

    def _task_equivalent(self, other):
        return self.wrapped_task._task_equivalent(other)


class FuncTask(Task):
    def __init__(self, func=None, equivalent=None, *a, **k):
        (super().__init__)(*a, **k)
        self._func = func
        self._equivalent = equivalent

    def _set_func(self, func):
        self._func = func

    def _get_func(self):
        return self._orig

    def do_update(self, timer):
        super().do_update(timer)
        action = self._func(timer)
        if not action or action == KILLED:
            self.kill()
        else:
            if action == PAUSED:
                self.pause()

    def _task_equivalent(self, other):
        return other in (self, self._func, self._equivalent)

    func = property(_get_func, _set_func)


class GeneratorTask(Task):
    class Param:
        delta = 0

    def __init__(self, generator=None, equivalent=None, *a, **k):
        (super().__init__)(*a, **k)
        self._param = GeneratorTask.Param()
        self._set_generator(generator)
        self._equivalent = equivalent

    def _set_generator(self, generator):
        self._orig = generator
        self._iter = generator(self._param)

    def _get_generator(self):
        return self._orig

    def do_update(self, delta):
        super().do_update(delta)
        self._param.delta = delta
        try:
            action = next(self._iter)
        except StopIteration:
            action = KILLED

        if not action or action == KILLED:
            self.kill()
        else:
            if action == PAUSED:
                self.pause()

    def _task_equivalent(self, other):
        return other in (self, self._orig, self._equivalent)

    generator = property(_get_generator, _set_generator)


class TaskGroup(Task):
    auto_kill = True
    auto_remove = True
    loop = False

    def __init__(self, tasks=None, auto_kill=None, auto_remove=None, loop=None, *a, **k):
        if tasks is None:
            tasks = []
        (super().__init__)(*a, **k)
        if auto_kill is not None:
            self.auto_kill = auto_kill
        if auto_remove is not None:
            self.auto_remove = auto_remove
        if loop is not None:
            self.loop = loop
        self._tasks = []
        for task in tasks:
            self.add(task)

    def clear(self):
        for t in self._tasks:
            t._set_parent(None)

        self._tasks = []
        super().clear()

    @depends(log_message=(const(print_message)), traceback=(const(traceback)))
    def do_update(self, timer, log_message=None, traceback=None):
        super().do_update(timer)
        for task in self._tasks:
            if not task.is_killed:
                try:
                    task.update(timer)
                except Exception:
                    task.kill()
                    log_message("Error when executing task")
                    traceback.print_exc()

        if self.auto_remove:
            self._tasks = remove_if(lambda t: t.is_killed, self._tasks)
        all_killed = len([t for t in self._tasks if t.is_killed]) == self.count
        if self.auto_kill and all_killed:
            self.kill()
        else:
            if self.loop:
                if all_killed:
                    self.restart()

    def add(self, task):
        task = totask(task)
        task._set_parent(self)
        self._tasks.append(task)
        if self.is_killed:
            super().restart()
        return task

    def remove(self, task):
        self._tasks.remove(task)
        task._set_parent(None)

    def find(self, task):
        return find_if(lambda t: t._task_equivalent(task), self._tasks)

    def restart(self):
        super().restart()
        for x in self._tasks:
            x.restart()

    @property
    def count(self):
        return len(self._tasks)


class WaitTask(Task):
    duration = 1.0

    def __init__(self, duration=None, *a, **k):
        (super().__init__)(*a, **k)
        if duration is not None:
            self.duration = duration
        self.remaining = self.duration

    def do_update(self, delta):
        super().do_update(delta)
        self.remaining -= delta
        if self.remaining <= 0:
            self.kill()
            self.remaining = 0

    def do_restart(self):
        self.remaining = self.duration


class DelayTask(Task):
    duration = 1

    def __init__(self, duration=None, *a, **k):
        (super().__init__)(*a, **k)
        if duration is not None:
            self.duration = duration
        self.remaining = self.duration

    def do_restart(self):
        self.remaining = self.duration

    def do_update(self, delta):
        super().do_update(delta)
        self.remaining -= 1
        if self.remaining <= 0:
            self.kill()
            self.remaining = 0


class TimerTask(WaitTask):
    def do_update(self, timer):
        super().do_update(timer)
        if self.remaining == 0:
            self.on_finish()
        else:
            self.on_tick()

    def on_tick(self):
        pass

    def on_finish(self):
        pass


class FadeTask(Task):
    def __init__(self, func=lambda x: x, duration=1.0, loop=False, init=False, *a, **k):
        (super().__init__)(*a, **k)
        self.func = func
        self.curr = 0.0
        self.loop = loop
        self.duration = duration
        if init:
            func(0.0)

    def do_update(self, delta):
        super().do_update(delta)
        self.func(old_div(self.curr, self.duration))
        self.curr += delta
        if self.curr >= self.duration:
            if self.loop:
                self.curr -= self.duration
            else:
                self.curr = self.duration
                self.kill()
                self.func(old_div(self.curr, self.duration))

    def do_restart(self):
        super().do_restart()
        self.curr = 0.0


class SequenceTask(Task):
    def __init__(self, tasks=None, *a, **k):
        if tasks is None:
            tasks = []
        (super().__init__)(*a, **k)
        self._tasks = tasks
        self._iter = iter(tasks)
        self._current = None
        self._advance_sequence()

    def _advance_sequence(self):
        try:
            self._current = next(self._iter)
        except StopIteration:
            self.kill()

    def do_update(self, delta):
        super().do_update(delta)
        if self._current is not None:
            self._current.update(delta)
            if self._current.is_killed:
                self._advance_sequence()

    def do_restart(self):
        for x in self._tasks:
            x.restart()

        self._iter = iter(self._tasks)
        self._advance_sequence()


class TimedCallbackTask(SequenceTask):
    _callback = nop

    def start(self, duration, callback):
        if duration is not None:
            self._tasks = [DelayTask(duration), FuncTask(self._call)]
            self._callback = callback or nop
        else:
            self._tasks = []
        self.restart()

    def _call(self, _time_expired):
        self.cancel()
        self._callback()

    def cancel(self):
        self.kill()


def totask(task):
    if not isinstance(task, Task):
        if not callable(task):
            raise TaskError("You can add either tasks or callables. " + str(task))
        task = FuncTask(func=task)
    return task


def generator(orig):
    equiv = [None]

    @functools.wraps(orig)
    def wrapper():
        return GeneratorTask(orig, equiv[0])

    equiv[0] = wrapper
    return wrapper


def sequence(*tasks):
    return SequenceTask(list(map(totask, tasks)))


def parallel(*tasks):
    return TaskGroup(tasks=tasks, auto_remove=False)


def loop(*tasks):
    return TaskGroup(tasks=tasks, auto_remove=False, auto_kill=False, loop=True)


wait = WaitTask
fade = FadeTask
delay = DelayTask


def invfade(f, *a, **k):
    return fade(lambda x: f(1.0 - x), *a, **k)


def linear(f, min, max, *a, **k):
    return fade(lambda x: f(linear_fn(min, max, x)), *a, **k)


try:
    import math

    def sinusoid(f, min=0.0, max=1.0, *a, **k):
        return fade(lambda x: f(min + (max - min) * math.sin(x * old_div(math.pi, 2.0))), *a, **k)


except ImportError:
    try:
        pass
    finally:
        err = None
        del err


def run(func, *a, **k):
    return FuncTask(
        lambda t: None if func(*a, **k) else None,
    )


def repeat(task):
    task = totask(task)
    task.kill = task.restart
    return WrapperTask(task)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/parameter_info.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 950 bytes
from ableton.v2.control_surface import ParameterInfo as ParameterInfoBase

from ..live import liveobj_valid


class ParameterInfo(ParameterInfoBase):
    def __init__(self, parameter=None, name=None, *a, **k):
        (super().__init__)(a, parameter=parameter, name=name, **k)
        if liveobj_valid(parameter):
            if name is not None:
                parameter.display_name = name

    @property
    def original_name(self):
        if liveobj_valid(self.parameter):
            return self.parameter.original_name
        return ""
